"""Various validation datapipes"""
from .check_equality import CheckGreaterThanOrEqualToIterDataPipe as CheckGreaterThanOrEqualTo
from .check_equality import CheckLessThanOrEqualToIterDataPipe as CheckLessThanOrEqualTo
from .check_equality import CheckNotEqualToIterDataPipe as CheckNotEqualTo
from .check_equality import CheckValueEqualToFractionIterDataPipe as CheckValueEqualToFraction
from .check_for_nans import CheckNaNsIterDataPipe as CheckNaNs
from .check_vars_and_dims import CheckVarsAndDimsIterDataPipe as CheckVarsAndDims



"""ECMWF provider loaders"""
import pandas as pd
import xarray as xr

from ocf_datapipes.load.nwp.providers.utils import open_zarr_paths


def open_ifs(zarr_path) -> xr.DataArray:
    """
    Opens the ECMWF IFS NWP data

    Args:
        zarr_path: Path to the zarr to open

    Returns:
        Xarray DataArray of the NWP data
    """
    # Open the data
    nwp = open_zarr_paths(zarr_path)
    dataVars = list(nwp.data_vars.keys())
    if len(dataVars) > 1:
        raise Exception("Too many TLDVs")
    else:
        dataVar = dataVars[0]
    ifs: xr.Dataset = nwp[dataVar]
    del nwp
    ifs = ifs.transpose("init_time", "step", "variable", "latitude", "longitude")
    ifs = ifs.rename(
        {
            "init_time": "init_time_utc",
            "variable": "channel",
        }
    )
    # Sanity checks.
    time = pd.DatetimeIndex(ifs.init_time_utc)
    assert time.is_unique
    assert time.is_monotonic_increasing
    return ifs



"""Open GFS Forecast data"""
import logging
from pathlib import Path
from typing import Union

import pandas as pd
import xarray as xr

_log = logging.getLogger(__name__)


def open_gfs(zarr_path: Union[Path, str]) -> xr.Dataset:
    """
    Opens GFS dataset

    Args:
        zarr_path: Path to Zarr(s) to open

    Returns:
        Xarray dataset of GFS Forecasts
    """

    _log.info("Loading NWP GFS data")

    if "*" in zarr_path:
        nwp = xr.open_mfdataset(zarr_path, engine="zarr", combine="time", chunks="auto")
    else:
        nwp = xr.load_dataset(zarr_path, engine="zarr", mode="r", chunks="auto")

    variables = list(nwp.keys())

    nwp = xr.concat([nwp[v] for v in variables], "channel")
    nwp = nwp.assign_coords(channel=variables)

    nwp = nwp.transpose("time", "step", "channel", "latitude", "longitude")
    nwp = nwp.rename({"time": "init_time_utc"})
    nwp = nwp.transpose("init_time_utc", "step", "channel", "latitude", "longitude")
    if "valid_time" in nwp.coords.keys():
        nwp = nwp.drop("valid_time")

    _log.debug("Interpolating hour 0 to NWP data")
    nwp_step0 = nwp.interp(step=[pd.Timedelta(hours=0)])
    nwp = xr.concat([nwp_step0, nwp], dim="step")
    nwp = nwp.resample(init_time_utc="60min").pad()
    nwp = nwp.resample(step="60min").pad()

    _log.debug(nwp)

    return nwp



"""Various utilites that didn't fit elsewhere"""
import logging
from pathlib import Path
from typing import Tuple, Union

import fsspec.asyn
import numpy as np
import numpy.typing as npt
import pandas as pd
import xarray as xr
from pandas.core.dtypes.common import is_datetime64_dtype
from pathy import Pathy

logger = logging.getLogger(__name__)


def datetime64_to_float(datetimes: np.ndarray, dtype=np.float64) -> np.ndarray:
    """
    Converts datetime64 to floats

    Args:
        datetimes: Array of datetimes
        dtype: Dtype to convert to

    Returns:
        Converted datetimes to floats
    """
    nums = datetimes.astype("datetime64[s]").astype(dtype)
    mask = np.isfinite(datetimes)
    return np.where(mask, nums, np.NaN)


def is_sorted(array: np.ndarray) -> bool:
    """Return True if array is sorted in ascending order."""
    # Adapted from https://stackoverflow.com/a/47004507/732596
    if len(array) == 0:
        return False
    return np.all(array[:-1] <= array[1:])


def searchsorted(a, v, side="left", assume_ascending=True):
    """Find indices where elements should be inserted to maintain order.

    Can be either sorted in ascending order or descending order

    Args:
        a: Target array
        v: Values to insert into `a`
        side: If ‘left’, the index of the first suitable location found is given. If ‘right’, return
            the last such index
        assume_ascending: Whether `a` is in ascending order, else assumed descending
    """
    if assume_ascending:
        return np.searchsorted(a, v, side=side)
    else:
        return np.searchsorted(-a, -v, side=side)


def check_path_exists(path: Union[str, Path]):
    """Raise a FileNotFoundError if `path` does not exist.

    `path` can include wildcards.
    """
    if not path:
        raise FileNotFoundError("Not a valid path!")
    filesystem = get_filesystem(path)
    if not filesystem.exists(path):
        # Maybe `path` includes a wildcard? So let's use `glob` to check.
        # Try `exists` before `glob` because `glob` might be slower.
        files = filesystem.glob(path)
        if len(files) == 0:
            raise FileNotFoundError(f"{path} does not exist!")


def get_filesystem(path: Union[str, Path]) -> fsspec.AbstractFileSystem:
    r"""Get the fsspect FileSystem from a path.

    For example, if `path` starts with `gs://` then return a gcsfs.GCSFileSystem.

    It is safe for `path` to include wildcards in the final filename.
    """
    path = Pathy(path)
    return fsspec.open(path.parent).fs


def set_fsspec_for_multiprocess() -> None:
    """
    Clear reference to the loop and thread.

    This is a nasty hack that was suggested but NOT recommended by the lead fsspec developer!

    This appears necessary otherwise gcsfs hangs when used after forking multiple worker processes.
    Only required for fsspec >= 0.9.0

    See:
    - https://github.com/fsspec/gcsfs/issues/379#issuecomment-839929801
    - https://github.com/fsspec/filesystem_spec/pull/963#issuecomment-1131709948

    TODO: Try deleting this two lines to make sure this is still relevant.
    """
    fsspec.asyn.iothread[0] = None
    fsspec.asyn.loop[0] = None
    fsspec.asyn._lock = None


def _trig_transform(values: np.ndarray, period: Union[float, int]) -> Tuple[np.ndarray, np.ndarray]:
    """
    Given a list of values and an upper limit on the values, compute trig decomposition.

    Args:
        values: ndarray of points in the range [0, period]
        period: period of the data

    Returns:
        Decomposition of values into sine and cosine of data with given period
    """

    return np.sin(values * 2 * np.pi / period), np.cos(values * 2 * np.pi / period)


def trigonometric_datetime_transformation(datetimes: npt.ArrayLike) -> np.ndarray:
    """
    Given an iterable of datetimes, returns a trigonometric decomposition on hour, day and month

    Args:
        datetimes: ArrayLike of datetime64 values

    Returns:
        Trigonometric decomposition of datetime into hourly, daily and
        monthly values.
    """
    assert is_datetime64_dtype(datetimes), "Data for Trig Decomposition must be np.datetime64 type"

    datetimes = pd.DatetimeIndex(datetimes)
    hour = datetimes.hour.values.reshape(-1, 1) + (datetimes.minute.values.reshape(-1, 1) / 60)
    day = datetimes.day.values.reshape(-1, 1)
    month = datetimes.month.values.reshape(-1, 1)

    sine_hour, cosine_hour = _trig_transform(hour, 24)
    sine_day, cosine_day = _trig_transform(day, 366)
    sine_month, cosine_month = _trig_transform(month, 12)

    return np.concatenate(
        [sine_month, cosine_month, sine_day, cosine_day, sine_hour, cosine_hour], axis=1
    )


def combine_to_single_dataset(dataset_dict: dict[str, xr.Dataset]) -> xr.Dataset:
    """
    Combine multiple datasets into a single dataset

    Args:
        dataset_dict: Dictionary of xr.Dataset objects to combine

    Returns:
        Combined dataset
    """
    # Flatten any NWP data
    dataset_dict = flatten_nwp_source_dict(dataset_dict, sep="-")

    # Convert all data_arrays to datasets
    new_dataset_dict = {}
    for key, datasets in dataset_dict.items():
        new_datasets = []
        for dataset in datasets:
            # Convert all coordinates float64 and int64 to float32 and int32
            dataset = dataset.assign_attrs(
                {key: str(value) for key, value in dataset.attrs.items()}
            )
            if isinstance(dataset, xr.DataArray):
                new_datasets.append(dataset.to_dataset(name=key))
            else:
                new_datasets.append(dataset)
            assert isinstance(new_datasets[-1], xr.Dataset)
        new_dataset_dict[key] = new_datasets
    # Prepend all coordinates and dimensions names with the key in the dataset_dict
    final_datasets_to_combined = []
    for key, datasets in new_dataset_dict.items():
        batched_datasets = []
        for dataset in datasets:
            dataset = dataset.rename(
                {dim: f"{key}__{dim}" for dim in dataset.dims if dim not in dataset.coords}
            )
            dataset = dataset.rename({coord: f"{key}__{coord}" for coord in dataset.coords})
            batched_datasets.append(dataset)
        # Merge all datasets with the same key
        # If NWP, then has init_time_utc and step, so do it off key__init_time_utc
        dataset = xr.concat(
            batched_datasets,
            dim=f"{key}__target_time_utc"
            if f"{key}__target_time_utc" in dataset.coords
            else f"{key}__time_utc",
        )
        # Serialize attributes to be JSON-seriaizable
        final_datasets_to_combined.append(dataset)
    # Combine all datasets, and append the list of datasets to the dataset_dict
    for f_dset in final_datasets_to_combined:
        assert isinstance(f_dset, xr.Dataset), f"Dataset is not an xr.Dataset, {type(f_dset)}"
    combined_dataset = xr.merge(final_datasets_to_combined)
    # Print all attrbutes of the combined dataset
    return combined_dataset


def uncombine_from_single_dataset(combined_dataset: xr.Dataset) -> dict[str, xr.DataArray]:
    """
    Uncombine a combined dataset

    Args:
        combined_dataset: The combined NetCDF dataset

    Returns:
        The uncombined datasets as a dict of xr.Datasets
    """
    # Split into datasets by splitting by the prefix added in combine_to_netcdf
    datasets = {}
    # Go through each data variable and split it into a dataset
    for key, dataset in combined_dataset.items():
        # If 'key_' doesn't exist in a dim or coordinate, remove it
        dataset_dims = list(dataset.coords)
        for dim in dataset_dims:
            if f"{key}__" not in dim:
                dataset: xr.DataArray = dataset.drop(dim)
        dataset = dataset.rename(
            {dim: dim.split(f"{key}__")[1] for dim in dataset.dims if dim not in dataset.coords}
        )
        dataset: xr.Dataset = dataset.rename(
            {coord: coord.split(f"{key}__")[1] for coord in dataset.coords}
        )
        # Split the dataset by the prefix
        datasets[key] = dataset

    # Unflatten any NWP data
    datasets = nest_nwp_source_dict(datasets, sep="-")
    return datasets


def flatten_nwp_source_dict(d: dict, sep: str = "/") -> dict:
    """Unnest a dictionary where the NWP values are nested under the key 'nwp'."""
    new_dict = {k: v for k, v in d.items() if k != "nwp"}
    if "nwp" in d:
        if isinstance(d["nwp"], dict):
            new_dict.update({f"nwp{sep}{k}": v for k, v in d["nwp"].items()})
        else:
            new_dict.update({"nwp": d["nwp"]})
    return new_dict


def nest_nwp_source_dict(d: dict, sep: str = "/") -> dict:
    """Re-nest a dictionary where the NWP values are nested under keys 'nwp/<key>'."""
    nwp_prefix = f"nwp{sep}"
    new_dict = {k: v for k, v in d.items() if not k.startswith(nwp_prefix)}
    nwp_keys = [k for k in d.keys() if k.startswith(nwp_prefix)]
    if len(nwp_keys) > 0:
        nwp_subdict = {k.removeprefix(nwp_prefix): d[k] for k in nwp_keys}
        new_dict["nwp"] = nwp_subdict
    return new_dict



"""UKV provider loaders"""
import numpy as np
import pandas as pd
import xarray as xr

from ocf_datapipes.load.nwp.providers.utils import open_zarr_paths


def open_ukv(zarr_path) -> xr.DataArray:
    """
    Opens the NWP data

    Args:
        zarr_path: Path to the zarr to open

    Returns:
        Xarray DataArray of the NWP data
    """
    # Open the data
    nwp = open_zarr_paths(zarr_path)
    ukv: xr.DataArray = nwp["UKV"]
    del nwp
    ukv = ukv.transpose("init_time", "step", "variable", "y", "x")
    ukv = ukv.rename(
        {
            "init_time": "init_time_utc",
            "variable": "channel",
            "y": "y_osgb",
            "x": "x_osgb",
        }
    )
    # y_osgb and x_osgb are int64 on disk.
    for coord_name in ("y_osgb", "x_osgb"):
        ukv[coord_name] = ukv[coord_name].astype(np.float32)
    # Sanity checks.
    assert ukv.y_osgb[0] > ukv.y_osgb[1], "UKV must run from top-to-bottom."
    time = pd.DatetimeIndex(ukv.init_time_utc)
    assert time.is_unique
    assert time.is_monotonic_increasing
    return ukv



"""Convert GSP to Numpy Batch"""
import logging

from torch.utils.data import IterDataPipe, functional_datapipe

from ocf_datapipes.batch import BatchKey, NumpyBatch
from ocf_datapipes.utils.utils import datetime64_to_float

logger = logging.getLogger(__name__)


@functional_datapipe("convert_gsp_to_numpy_batch")
class ConvertGSPToNumpyBatchIterDataPipe(IterDataPipe):
    """Convert GSP Xarray to NumpyBatch"""

    def __init__(self, source_datapipe: IterDataPipe):
        """
        Convert GSP Xarray to NumpyBatch object

        Args:
            source_datapipe: Datapipe emitting GSP Xarray object
        """
        super().__init__()
        self.source_datapipe = source_datapipe

    def __iter__(self) -> NumpyBatch:
        """Convert from Xarray to NumpyBatch"""
        logger.debug("Converting GSP to numpy to batch")
        for xr_data in self.source_datapipe:
            example: NumpyBatch = {
                BatchKey.gsp: xr_data.values,
                BatchKey.gsp_t0_idx: xr_data.attrs["t0_idx"],
                BatchKey.gsp_id: xr_data.gsp_id.values,
                BatchKey.gsp_nominal_capacity_mwp: xr_data.isel(time_utc=0)[
                    "nominal_capacity_mwp"
                ].values,
                BatchKey.gsp_effective_capacity_mwp: (
                    xr_data.isel(time_utc=0)["effective_capacity_mwp"].values
                ),
                BatchKey.gsp_time_utc: datetime64_to_float(xr_data["time_utc"].values),
            }

            # Coordinates
            for batch_key, dataset_key in (
                (BatchKey.gsp_y_osgb, "y_osgb"),
                (BatchKey.gsp_x_osgb, "x_osgb"),
            ):
                if dataset_key in xr_data.coords.keys():
                    values = xr_data[dataset_key].values
                    # Expand dims so AddFourierSpaceTime works!
                    example[batch_key] = values  # np.expand_dims(values, axis=1)

            yield example



"""Merge individual examples into a batch"""
import logging
from typing import Sequence, Union

import numpy as np
from torch.utils.data import IterDataPipe, functional_datapipe

from ocf_datapipes.batch import BatchKey, NumpyBatch, NWPBatchKey, NWPNumpyBatch

logger = logging.getLogger(__name__)


def _key_is_constant(batch_key):
    is_constant = batch_key.name.endswith("t0_idx") or batch_key == NWPBatchKey.nwp_channel_names
    return is_constant


def stack_data_list(
    data_list: Sequence,
    batch_key: Union[BatchKey, NWPBatchKey],
):
    """How to combine data entries for each key

    See also: `extract_sample_from_batch()` for opposite
    """
    if _key_is_constant(batch_key):
        # These are always the same for all examples.
        return data_list[0]
    try:
        return np.stack(data_list)
    except Exception as e:
        logger.debug(f"Could not stack the following shapes together, ({batch_key})")
        shapes = [example.shape for example in data_list]
        logger.debug(shapes)
        logger.error(e)
        raise e


def extract_sample_from_batch(
    data,
    batch_key: Union[BatchKey, NWPBatchKey],
    index_num: int,
):
    """How to extract data entries for each key

    See also: `stack_data_list()` for opposite
    """
    if _key_is_constant(batch_key):
        # These are always the same for all examples.
        return data
    else:
        return data[index_num]


def stack_np_examples_into_batch(dict_list: Sequence[NumpyBatch]) -> NumpyBatch:
    """
    Stacks Numpy examples into a batch

    See also: `unstack_np_batch_into_examples()` for opposite

    Args:
        dict_list: A list of dict-like Numpy examples to stack

    Returns:
        The stacked NumpyBatch object
    """
    batch: NumpyBatch = {}

    batch_keys = list(dict_list[0].keys())

    for batch_key in batch_keys:
        # NWP is nested so treat separately
        if batch_key == BatchKey.nwp:
            nwp_batch: dict[str, NWPNumpyBatch] = {}

            # Unpack source keys
            nwp_sources = list(dict_list[0][BatchKey.nwp].keys())

            for nwp_source in nwp_sources:
                # Keys can be different for different NWPs
                nwp_batch_keys = list(dict_list[0][BatchKey.nwp][nwp_source].keys())

                nwp_source_batch: NWPNumpyBatch = {}
                for nwp_batch_key in nwp_batch_keys:
                    nwp_source_batch[nwp_batch_key] = stack_data_list(
                        [d[BatchKey.nwp][nwp_source][nwp_batch_key] for d in dict_list],
                        nwp_batch_key,
                    )

                nwp_batch[nwp_source] = nwp_source_batch

            batch[BatchKey.nwp] = nwp_batch

        else:
            batch[batch_key] = stack_data_list(
                [d[batch_key] for d in dict_list],
                batch_key,
            )

    return batch


def unstack_np_batch_into_examples(batch: NumpyBatch):
    """Splits a single batch into samples.

    Note:
    This can be really useful when using presaved batches, so you can split the samples, reshuffle,
    and recombine into batches. This means batches can be rebatched each epoch.

    See also: `stack_np_examples_into_batch()` for opposite
    """
    batch_keys = list(batch.keys())

    # Look at a non-constant key and find batch_size. Trickier if key is NWP
    non_constant_key = next(filter(lambda x: not _key_is_constant(x), batch_keys))
    if non_constant_key == BatchKey.nwp:
        # NWP is nested so treat separately
        nwp_source = next(iter(batch[BatchKey.nwp].keys()))
        non_constant_nwp_key = next(
            filter(
                lambda x: not _key_is_constant(x),
                list(batch[BatchKey.nwp][nwp_source].keys()),
            )
        )
        batch_size = batch[BatchKey.nwp][nwp_source][non_constant_nwp_key].shape[0]
    else:
        batch_size = batch[non_constant_key].shape[0]

    # Loop through and split the batch into samples
    samples = []
    for i in range(batch_size):
        sample: NumpyBatch = {}

        for key in batch_keys:
            # NWP is nested so treat separately
            if key == BatchKey.nwp:
                nwp_batch: dict[str, NWPNumpyBatch] = {}

                # Unpack keys
                nwp_sources = list(batch[BatchKey.nwp].keys())

                for nwp_source in nwp_sources:
                    nwp_source_batch: NWPNumpyBatch = {}

                    # Keys can be different for each NWP source
                    nwp_keys = list(batch[BatchKey.nwp][nwp_source].keys())

                    for nwp_key in nwp_keys:
                        nwp_source_batch[nwp_key] = extract_sample_from_batch(
                            batch[BatchKey.nwp][nwp_source][nwp_key],
                            batch_key=nwp_key,
                            index_num=i,
                        )

                    nwp_batch[nwp_source] = nwp_source_batch

                sample[BatchKey.nwp] = nwp_batch

            else:
                sample[key] = extract_sample_from_batch(
                    batch[key],
                    batch_key=key,
                    index_num=i,
                )

        samples += [sample]
    return samples


@functional_datapipe("merge_numpy_batch")
class MergeNumpyBatchIterDataPipe(IterDataPipe):
    """Merge list of individual examples into a batch"""

    def __init__(self, source_datapipe: IterDataPipe):
        """
        Merge list of individual examples into a batch

        Args:
            source_datapipe: Datapipe of yielding lists of numpybatch examples
        """
        self.source_datapipe = source_datapipe

    def __iter__(self) -> NumpyBatch:
        """Merge list of individual examples into a batch"""
        logger.debug("Merging numpy batch")
        for examples_list in self.source_datapipe:
            yield stack_np_examples_into_batch(examples_list)


# TODO: Is this needed anymore? Instead we can do either of:
#  - `dp.batch(batch_size).merge_numpy_batch()`
#  - `dp.batch(batch_size).map(stack_np_examples_into_batch)`
@functional_datapipe("merge_numpy_examples_to_batch")
class MergeNumpyExamplesToBatchIterDataPipe(IterDataPipe):
    """Merge individual examples into a batch"""

    def __init__(self, source_datapipe: IterDataPipe, n_examples_per_batch: int):
        """
        Merge individual examples into a batch

        Args:
            source_datapipe: Datapipe of NumpyBatch data
            n_examples_per_batch: Number of examples per batch
        """
        self.source_datapipe = source_datapipe
        self.n_examples_per_batch = n_examples_per_batch

    def __iter__(self) -> NumpyBatch:
        """Merge individual examples into a batch"""
        np_examples = []
        logger.debug("Merging numpy batch")
        for np_batch in self.source_datapipe:
            np_examples.append(np_batch)
            if len(np_examples) == self.n_examples_per_batch:
                yield stack_np_examples_into_batch(np_examples)
                np_examples = []



"""Satellite loader"""
import logging
import subprocess
from pathlib import Path
from typing import Union

import dask
import pandas as pd
import xarray as xr
from ocf_blosc2 import Blosc2  # noqa: F401
from torch.utils.data import IterDataPipe, functional_datapipe

_log = logging.getLogger(__name__)


def _get_single_sat_data(zarr_path: Union[Path, str]) -> xr.DataArray:
    """Helper function to open a zarr from either local or GCP path.

    The local or GCP path may contain wildcard matching (*)

    Args:
        zarr_path: Path to zarr file
    """

    # These kwargs are used if zarr path contains "*"
    openmf_kwargs = dict(
        engine="zarr",
        concat_dim="time",
        combine="nested",
        chunks="auto",
        join="override",
    )

    # Need to generate list of files if using GCP bucket storage
    if "gs://" in str(zarr_path) and "*" in str(zarr_path):
        result_string = subprocess.run(
            f"gsutil ls -d {zarr_path}".split(" "), stdout=subprocess.PIPE
        ).stdout.decode("utf-8")
        files = result_string.splitlines()

        dataset = xr.open_mfdataset(files, **openmf_kwargs)

    elif "*" in str(zarr_path):  # Multi-file dataset
        dataset = xr.open_mfdataset(zarr_path, **openmf_kwargs)
    else:
        dataset = xr.open_dataset(zarr_path, engine="zarr", chunks="auto")
    dataset = dataset.drop_duplicates("time").sortby("time")

    return dataset


def open_sat_data(zarr_path: Union[Path, str, list[Path], list[str]]) -> xr.DataArray:
    """Lazily opens the Zarr store.

    Args:
      zarr_path: Cloud URL or local path pattern, or list of these. If GCS URL, it must start with
          'gs://'.

    Example:
        With wild cards and GCS path:
        ```
        zarr_paths = [
            "gs://bucket/2020_nonhrv_split_*.zarr",
            "gs://bucket/2019_nonhrv_split_*.zarr",
        ]
        ds = open_sat_data(zarr_paths)
        ```
        Without wild cards and with local path:
        ```
        zarr_paths = [
            "/data/2020_nonhrv.zarr",
            "/data/2019_nonhrv.zarr",
        ]
        ds = open_sat_data(zarr_paths)
        ```
    """

    # Silence the warning about large chunks.
    # Alternatively, we could set this to True, but that slows down loading a Satellite batch
    # from 8 seconds to 50 seconds!
    dask.config.set({"array.slicing.split_large_chunks": False})

    if isinstance(zarr_path, (list, tuple)):
        message_files_list = "\n - " + "\n - ".join([str(s) for s in zarr_path])
        _log.info(f"Opening satellite data: {message_files_list}")
        dataset = xr.combine_nested(
            [_get_single_sat_data(path) for path in zarr_path],
            concat_dim="time",
            combine_attrs="override",
            join="override",
        )
    else:
        _log.info(f"Opening satellite data: {zarr_path}")
        dataset = _get_single_sat_data(zarr_path)

    # Remove data coordinate dimensions if they exist
    if "x_geostationary_coordinates" in dataset:
        del dataset["x_geostationary_coordinates"]
        del dataset["y_geostationary_coordinates"]

    # Rename
    # These renamings will no longer be necessary when the Zarr uses the 'correct' names,
    # see https://github.com/openclimatefix/Satip/issues/66
    if "variable" in dataset:
        dataset = dataset.rename({"variable": "channel"})
    if "channels" in dataset:
        dataset = dataset.rename({"channels": "channel"})
    elif "channel" not in dataset:
        # This is HRV version 3, which doesn't have a channels dim.  So add one.
        dataset = dataset.expand_dims(dim={"channel": ["HRV"]}, axis=1)

    # Rename coords to be more explicit about exactly what some coordinates hold:
    # Note that `rename` renames *both* the coordinates and dimensions, and keeps
    # the connection between the dims and coordinates, so we don't have to manually
    # use `data_array.set_index()`.
    dataset = dataset.rename(
        {
            "time": "time_utc",
        }
    )
    if "y" in dataset.coords.keys():
        dataset = dataset.rename(
            {
                "y": "y_geostationary",
            }
        )

    if "x" in dataset.coords.keys():
        dataset = dataset.rename(
            {
                "x": "x_geostationary",
            }
        )

    # Flip coordinates to top-left first
    if dataset.y_geostationary[0] < dataset.y_geostationary[-1]:
        dataset = dataset.reindex(y_geostationary=dataset.y_geostationary[::-1])
    if dataset.x_geostationary[0] > dataset.x_geostationary[-1]:
        dataset = dataset.reindex(x_geostationary=dataset.x_geostationary[::-1])

    data_array = dataset["data"]
    del dataset

    # Ensure the y and x coords are in the right order (top-left first):
    assert data_array.y_geostationary[0] > data_array.y_geostationary[-1]
    assert data_array.x_geostationary[0] < data_array.x_geostationary[-1]
    if "y_osgb" in data_array.dims:
        assert data_array.y_osgb[0, 0] > data_array.y_osgb[-1, 0]
        assert data_array.x_osgb[0, 0] < data_array.x_osgb[0, -1]

    # Sanity checks!
    data_array = data_array.transpose("time_utc", "channel", "y_geostationary", "x_geostationary")
    assert data_array.dims == ("time_utc", "channel", "y_geostationary", "x_geostationary")
    datetime_index = pd.DatetimeIndex(data_array.time_utc)
    assert datetime_index.is_unique
    assert datetime_index.is_monotonic_increasing
    # Satellite datetimes can sometimes be 04, 09, minutes past the hour, or other slight offsets.
    # These slight offsets will break downstream code, which expects satellite data to be at
    # exactly 5 minutes past the hour.
    assert (datetime_index == datetime_index.round("5min")).all()

    _log.info("Opened satellite data")

    return data_array


@functional_datapipe("open_satellite")
class OpenSatelliteIterDataPipe(IterDataPipe):
    """Open Satellite Zarr"""

    def __init__(self, zarr_path: Union[Path, str]):
        """
        Opens the satellite Zarr

        Args:
            zarr_path: path to the zarr file
        """
        self.zarr_path = zarr_path
        super().__init__()

    def __iter__(self) -> xr.DataArray:
        """Open the Zarr file"""
        data: xr.DataArray = open_sat_data(zarr_path=self.zarr_path)
        while True:
            yield data



"""Load Topographic data"""
from pathlib import Path
from typing import Union

import rioxarray
from torch.utils.data import IterDataPipe, functional_datapipe


@functional_datapipe("open_topography")
class OpenTopographyIterDataPipe(IterDataPipe):
    """Load topographic data"""

    def __init__(self, topo_filename: Union[Path, str]):
        """
        Load topographic data

        Args:
            topo_filename: Filename to open
        """
        self.topo_filename = topo_filename

    def __iter__(self):
        """Load topo data and return it"""
        topo = rioxarray.open_rasterio(
            filename=self.topo_filename, parse_coordinates=True, masked=True
        )

        # `band` and `spatial_ref` don't appear to hold any useful info. So get rid of them:
        topo = topo.isel(band=0)
        topo = topo.drop_vars(["spatial_ref", "band"])

        # Use our standard naming:
        topo = topo.rename({"x": "x_osgb", "y": "y_osgb"})

        while True:
            yield topo



"""Convert NWP data to the target time with dropout"""
import logging
from datetime import timedelta
from typing import List, Optional, Union

import numpy as np
import pandas as pd
import xarray as xr
from torch.utils.data import IterDataPipe, functional_datapipe

logger = logging.getLogger(__name__)


@functional_datapipe("select_time_slice_nwp")
class SelectTimeSliceNWPIterDataPipe(IterDataPipe):
    """Convert NWP Xarray dataset to use target time as indexer"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        t0_datapipe: IterDataPipe,
        sample_period_duration: timedelta,
        history_duration: timedelta,
        forecast_duration: timedelta,
        dropout_timedeltas: Optional[List[timedelta]] = None,
        dropout_frac: Optional[float] = 0,
    ):
        """Convert NWP Xarray dataset to use target time as indexer

        Args:
            source_datapipe: Datapipe emitting an Xarray Dataset with step and init_time_utc
                indexers.
            t0_datapipe: Datapipe emitting t0 times for indexing off of choosing the closest
                previous init_time_utc.
            sample_period_duration: How long the sampling period is.
            history_duration: How long the history time should cover.
            forecast_duration: How long the forecast time should cover.
            dropout_timedeltas: List of timedeltas. We randonly select the delay for each NWP
                forecast from this list. These should be negative timedeltas w.r.t time t0.
            dropout_frac: Fraction of samples subject to dropout
        """
        self.source_datapipe = source_datapipe
        self.t0_datapipe = t0_datapipe
        self.sample_period_duration = sample_period_duration
        self.history_duration = history_duration
        self.forecast_duration = forecast_duration
        self.dropout_timedeltas = dropout_timedeltas
        self.dropout_frac = dropout_frac
        if dropout_timedeltas is not None:
            assert all(
                [t < timedelta(minutes=0) for t in dropout_timedeltas]
            ), "dropout timedeltas must be negative"
            assert len(dropout_timedeltas) >= 1
        assert 0 <= dropout_frac <= 1
        self._consider_dropout = (dropout_timedeltas is not None) and dropout_frac > 0

    def __iter__(self) -> Union[xr.DataArray, xr.Dataset]:
        """Iterate through both datapipes and convert Xarray dataset"""

        xr_data = next(iter(self.source_datapipe))

        for t0 in self.t0_datapipe:
            t0 = pd.Timestamp(t0)
            start_dt = t0 - self.history_duration
            end_dt = t0 + self.forecast_duration

            target_times = pd.date_range(
                start_dt.ceil(self.sample_period_duration),
                end_dt.ceil(self.sample_period_duration),
                freq=self.sample_period_duration,
            )

            # Maybe apply NWP dropout
            if self._consider_dropout and (np.random.uniform() < self.dropout_frac):
                dt = np.random.choice(self.dropout_timedeltas)
                t0_available = t0 + dt
            else:
                t0_available = t0

            # Forecasts made up to and including t0
            xr_available = xr_data.sel(init_time_utc=slice(None, t0_available))

            init_times = xr_available.sel(
                init_time_utc=target_times,
                method="ffill",  # forward fill from init times to target times
            ).init_time_utc.values

            steps = target_times - init_times

            # We want one timestep for each target_time_hourly (obviously!) If we simply do
            # nwp.sel(init_time=init_times, step=steps) then we'll get the *product* of
            # init_times and steps, which is not what # we want! Instead, we use xarray's
            # vectorized-indexing mode by using a DataArray indexer.  See the last example here:
            # https://docs.xarray.dev/en/latest/user-guide/indexing.html#more-advanced-indexing
            coords = {"target_time_utc": target_times}
            init_time_indexer = xr.DataArray(init_times, coords=coords)
            step_indexer = xr.DataArray(steps, coords=coords)
            xr_sel = xr_available.sel(step=step_indexer, init_time_utc=init_time_indexer)

            yield xr_sel



"""Select the t0 time for training"""
import logging

import numpy as np
import pandas as pd
from torch.utils.data import IterDataPipe, functional_datapipe

logger = logging.getLogger(__name__)


@functional_datapipe("pick_t0_times")
class PickT0TimesIterDataPipe(IterDataPipe):
    """Picks random t0 times from a dataset"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        dim_name: str = "time_utc",
    ):
        """
        Picks random t0 times from a dataset

        Args:
            source_datapipe: Datapipe emitting Xarray objects
            dim_name: The time dimension name to use
        """
        self.source_datapipe = source_datapipe
        self.dim_name = dim_name

    def __iter__(self) -> pd.Timestamp:
        """Get the latest timestamp and return it"""
        for xr_data in self.source_datapipe:
            logger.debug(f"Selecting t0 from {len(xr_data[self.dim_name])} datetimes")

            if len(xr_data[self.dim_name].values) == 0:
                raise Exception("There are no values to get t0 from")
            t0 = np.random.choice(xr_data[self.dim_name].values)
            logger.debug(f"t0 will be {t0}")

            yield t0



import os
import tempfile
from datetime import datetime, timedelta
import uuid


import numpy as np
import pandas as pd
import pytest
import xarray as xr
from nowcasting_datamodel.connection import DatabaseConnection
from nowcasting_datamodel.models import (
    Base_PV,
    GSPYield,
    Location,
    LocationSQL,
    PVSystem,
    PVSystemSQL,
    PVYield,
    pv_output,
    solar_sheffield_passiv,
)

import ocf_datapipes
from ocf_datapipes.config.load import load_yaml_configuration
from ocf_datapipes.config.model import PV, PVFiles
from ocf_datapipes.config.save import save_yaml_configuration
from ocf_datapipes.load import (
    OpenGSP,
    OpenNWP,
    OpenPVFromNetCDF,
    OpenSatellite,
    OpenTopography,
)

xr.set_options(keep_attrs=True)

# This path is used both here and in tests in deeper test directories
# Make two ways to easily access it
# Weirdly using pytest.fixture(autouse=True) wasn't working - TODO?
_top_test_directory = os.path.dirname(os.path.realpath(__file__))


@pytest.fixture()
def top_test_directory():
    return _top_test_directory


@pytest.fixture()
def sat_hrv_datapipe():
    filename = _top_test_directory + "/data/hrv_sat_data.zarr"
    return OpenSatellite(zarr_path=filename)


@pytest.fixture()
def sat_datapipe():
    filename = f"{_top_test_directory}/data/sat_data.zarr"
    # The saved data is scaled from 0-1024. Now we use data scaled from 0-1
    # Rescale here for subsequent tests
    return OpenSatellite(zarr_path=filename).map(lambda da: da / 1024)


@pytest.fixture()
def sat_15_datapipe():
    filename = f"{_top_test_directory}/data/sat_data_15.zarr"
    # The saved data is scaled from 0-1024. Now we use data scaled from 0-1
    # Rescale here for subsequent tests
    return OpenSatellite(zarr_path=filename).map(_sat_rescale)


@pytest.fixture()
def topo_datapipe():
    filename = f"{_top_test_directory}/data/europe_dem_2km_osgb.tif"
    return OpenTopography(topo_filename=filename)


@pytest.fixture()
def nwp_datapipe():
    filename = f"{_top_test_directory}/data/nwp_data/test.zarr"
    return OpenNWP(zarr_path=filename)


@pytest.fixture()
def icon_eu_datapipe():
    filename = f"{_top_test_directory}/data/icon_eu.zarr"
    return OpenNWP(zarr_path=filename, provider="icon-eu")


@pytest.fixture()
def icon_global_datapipe():
    filename = f"{_top_test_directory}/data/icon_global.zarr"
    return OpenNWP(zarr_path=filename, provider="icon-global")


@pytest.fixture()
def passiv_datapipe():
    filename = f"{_top_test_directory}/data/pv/passiv/test.nc"
    filename_metadata = f"{_top_test_directory}/data/pv/passiv/UK_PV_metadata.csv"

    pv = PV()
    pv_file = PVFiles(
        pv_filename=str(filename),
        pv_metadata_filename=str(filename_metadata),
        label="solar_sheffield_passiv",
    )
    pv.pv_files_groups = [pv_file]

    return OpenPVFromNetCDF(pv=pv)


@pytest.fixture()
def pvoutput_datapipe():
    filename = f"{_top_test_directory}/data/pv/pvoutput/test.nc"
    filename_metadata = f"{_top_test_directory}/data/pv/pvoutput/UK_PV_metadata.csv"

    pv = PV()
    pv_file = PVFiles(
        pv_filename=str(filename),
        pv_metadata_filename=str(filename_metadata),
        label="pvoutput.org",
    )
    pv.pv_files_groups = [pv_file]

    return OpenPVFromNetCDF(pv=pv)


@pytest.fixture()
def gsp_datapipe():
    filename = f"{_top_test_directory}/data/gsp/test.zarr"
    return OpenGSP(gsp_pv_power_zarr_path=filename)


@pytest.fixture
def pv_system_db_data():
    # Create generation data
    n_systems = 10

    t0 = pd.Timestamp.now().floor("5min")
    datetimes = pd.date_range(t0 - timedelta(minutes=120), t0, freq="5min")
    site_uuids = [str(uuid.uuid4()) for _ in range(n_systems)]

    data = np.zeros((len(datetimes), n_systems))

    # Make data a nice sinusoidal curve
    data[:] = (
        0.5
        * (1 - np.cos((datetimes.hour + datetimes.minute / 60) / 24 * 2 * np.pi).values)[:, None]
    )

    # Chuck in some nan values
    data[:, 1] = np.nan
    data[-5:, 2] = np.nan
    data[::3, 3] = np.nan

    da = xr.DataArray(
        data,
        coords=(
            ("end_utc", datetimes),
            ("site_uuid", site_uuids),
        ),
    )

    # Reshape for tabular database
    df_gen = da.to_dataframe("generation_power_kw").reset_index()
    df_gen["start_utc"] = df_gen["end_utc"] - timedelta(minutes=5)

    # Create metadata
    df_meta = pd.DataFrame(
        dict(
            site_uuid=site_uuids,
            orientation=np.random.uniform(0, 360, n_systems),
            tilt=np.random.uniform(0, 90, n_systems),
            longitude=np.random.uniform(-3.07, 0.59, n_systems),
            latitude=np.random.uniform(51.56, 52.89, n_systems),
            capacity_kw=np.random.uniform(1, 5, n_systems),
            ml_id=np.arange(n_systems),
        )
    )

    return df_gen, df_meta


@pytest.fixture
def db_connection(pv_system_db_data):
    """Create data connection"""

    with tempfile.NamedTemporaryFile(suffix=".db") as temp:
        url = f"sqlite:///{temp.name}"
        os.environ["DB_URL_PV"] = url
        os.environ["DB_URL"] = url

        connection = DatabaseConnection(url=url, base=Base_PV, echo=False)
        from nowcasting_datamodel.models import (
            GSPYieldSQL,
            LocationSQL,
            PVSystemSQL,
            PVYieldSQL,
        )

        for table in [PVYieldSQL, PVSystemSQL, GSPYieldSQL, LocationSQL]:
            table.__table__.create(connection.engine)

        # Create and populate pvsites tables
        df_gen, df_meta = pv_system_db_data
        with connection.engine.connect() as conn:
            df_gen.to_sql(name="generation", con=conn, index=False)
            df_meta.to_sql(name="sites", con=conn, index=False)

        yield connection


@pytest.fixture(scope="function", autouse=True)
def db_session(db_connection):
    """Creates a new database session for a test."""

    connection = db_connection.engine.connect()
    t = connection.begin()

    with db_connection.get_session() as s:
        s.begin()
        yield s
        s.rollback()

    t.rollback()
    connection.close()


@pytest.fixture()
def pv_yields_and_systems(db_session):
    """Create pv yields and systems

    Pv systems: Two systems
    PV yields:
        For system 1, pv yields from 4 to 10 at 5 minutes. Last one at 09.55
        For system 2: 1 pv yield at 04.00
    """

    pv_system_sql_1: PVSystemSQL = PVSystem(
        pv_system_id=1,
        provider="pvoutput.org",
        status_interval_minutes=5,
        longitude=0,
        latitude=55,
        ml_capacity_kw=123,
    ).to_orm()
    pv_system_sql_1_ss: PVSystemSQL = PVSystem(
        pv_system_id=1,
        provider=solar_sheffield_passiv,
        status_interval_minutes=5,
        longitude=0,
        latitude=57,
        ml_capacity_kw=124,
    ).to_orm()
    pv_system_sql_2: PVSystemSQL = PVSystem(
        pv_system_id=2,
        provider="pvoutput.org",
        status_interval_minutes=5,
        longitude=0,
        latitude=56,
        ml_capacity_kw=124,
    ).to_orm()
    pv_system_sql_3: PVSystemSQL = PVSystem(
        pv_system_id=3,
        provider=pv_output,
        status_interval_minutes=5,
        longitude=0,
        latitude=57,
        ml_capacity_kw=124,
    ).to_orm()

    pv_yield_sqls = []
    for hour in range(4, 10):
        for minute in range(0, 60, 5):
            pv_yield_1 = PVYield(
                datetime_utc=datetime(2022, 1, 1, hour, minute),
                solar_generation_kw=hour + minute / 100,
            ).to_orm()
            pv_yield_1.pv_system = pv_system_sql_1
            pv_yield_sqls.append(pv_yield_1)

            pv_yield_1_ss = PVYield(
                datetime_utc=datetime(2022, 1, 1, hour, minute),
                solar_generation_kw=hour + minute / 100,
            ).to_orm()
            pv_yield_1_ss.pv_system = pv_system_sql_1_ss
            pv_yield_sqls.append(pv_yield_1_ss)

    # pv system with gaps every 5 mins
    for minutes in [0, 10, 20, 30]:
        pv_yield_4 = PVYield(
            datetime_utc=datetime(2022, 1, 1, 4) + timedelta(minutes=minutes),
            solar_generation_kw=4,
        ).to_orm()
        pv_yield_4.pv_system = pv_system_sql_2
        pv_yield_sqls.append(pv_yield_4)

    # add a system with only on pv yield
    pv_yield_5 = PVYield(
        datetime_utc=datetime(2022, 1, 1, 4) + timedelta(minutes=minutes),
        solar_generation_kw=4,
    ).to_orm()
    pv_yield_5.pv_system = pv_system_sql_3
    pv_yield_sqls.append(pv_yield_5)

    # add to database
    db_session.add_all(pv_yield_sqls)
    db_session.add(pv_system_sql_1)
    db_session.add(pv_system_sql_2)

    db_session.commit()

    return {
        "pv_yields": pv_yield_sqls,
        "pv_systems": [pv_system_sql_1, pv_system_sql_2],
    }


@pytest.fixture()
def gsp_yields(db_session):
    """Make fake GSP data"""

    gsps = list(range(0, 4)) + [18, 317]

    gsp_yield_sqls = []
    for gsp_id in gsps:
        gsp_sql_1: LocationSQL = Location(
            gsp_id=gsp_id, label=f"GSP_{gsp_id}", installed_capacity_mw=1
        ).to_orm()

        for hour in range(0, 8):
            for minute in range(0, 60, 30):
                gsp_yield_1 = GSPYield(
                    datetime_utc=datetime(2022, 1, 1, hour, minute),
                    solar_generation_kw=hour + minute,
                    capacity_mwp=1,
                )
                gsp_yield_1_sql = gsp_yield_1.to_orm()
                gsp_yield_1_sql.location = gsp_sql_1
                gsp_yield_sqls.append(gsp_yield_1_sql)

    # add to database
    db_session.add_all(gsp_yield_sqls)
    db_session.commit()

    return {
        "gsp_yields": gsp_yield_sqls,
        "gsp_systems": [gsp_sql_1],
    }


@pytest.fixture()
def pv_xarray_data():
    datetimes = pd.date_range("2022-09-01 00:00", "2022-09-08 00:00", freq="5min")
    pv_system_ids = (np.arange(10) + 9905).astype(str)

    data = np.full((len(datetimes), len(pv_system_ids)), fill_value=9.1)
    da = xr.DataArray(
        data,
        coords=(
            ("datetime", datetimes),
            ("pv_system_id", pv_system_ids),
        ),
    )

    da = da.where(da.datetime.dt.hour < 21, other=0)
    da = da.where(da.datetime.dt.hour > 3, other=0)

    da.isel(datetime=slice(0, 3), pv_system_id=0).values[:] = np.nan
    return da


@pytest.fixture()
def pv_netcdf_file(pv_xarray_data):
    """Create a netcdf file with PV data with the following dimensions

    - datetime
    - pv_system_id
    """

    ds = pv_xarray_data.to_dataset(dim="pv_system_id")

    with tempfile.TemporaryDirectory() as tmpdir:
        filename = tmpdir + "data.nc"
        ds.to_netcdf(filename, engine="h5netcdf")
        yield filename


@pytest.fixture()
def pv_parquet_file(pv_xarray_data):
    """Create a parquet file with PV data with the following columns

    Columns
    - timestamp
    - ss_id
    - generation_wh
    """

    # Convert to watt-hours energy for each 5-minute step
    da = pv_xarray_data / 12

    # Flatten into DataFrame and rename
    data_df = da.to_dataframe("generation_wh").reset_index(level=[0, 1])
    data_df = data_df.rename(dict(datetime="timestamp", pv_system_id="ss_id"), axis=1)

    with tempfile.TemporaryDirectory() as tmpdir:
        filename = tmpdir + "/data.parquet"
        data_df.to_parquet(filename, engine="fastparquet")
        yield filename


@pytest.fixture()
def gsp_zarr_file():
    """GSP zarr file"""

    date = datetime(2022, 9, 1)
    days = 7
    ids = np.array(range(0, 10))
    datetime_gmt = pd.to_datetime([date + timedelta(minutes=30 * i) for i in range(0, days * 24)])

    coords = (
        ("datetime_gmt", datetime_gmt),
        ("gsp_id", ids),
    )

    generation_mw = xr.DataArray(
        np.random.uniform(
            0,
            200,
            size=(7 * 24, len(ids)),
        ),
        coords=coords,
        name="generation_mw",
    )

    installedcapacity_mwp = xr.DataArray(
        np.random.uniform(
            0,
            200,
            size=(7 * 24, len(ids)),
        ),
        coords=coords,
        name="installedcapacity_mwp",
    )

    capacity_mwp = xr.DataArray(
        np.random.uniform(
            0,
            200,
            size=(7 * 24, len(ids)),
        ),
        coords=coords,
        name="capacity_mwp",
    )

    generation_mw = generation_mw.to_dataset(name="generation_mw")
    generation_mw = generation_mw.merge(installedcapacity_mwp)
    generation_mw = generation_mw.merge(capacity_mwp)
    with tempfile.TemporaryDirectory() as tmpdir:
        filename = tmpdir + "/gsp.zarr"
        generation_mw.to_zarr(filename)

        yield filename


@pytest.fixture()
def nwp_data_with_id_filename():
    """Create xarray netcdf file for NWP data

    Variables
    - init_time
    - step
    - variables
    - id
    """

    init_times = pd.date_range(start=datetime(2022, 9, 1), freq="60min", periods=24 * 7)
    steps = [timedelta(minutes=60 * i) for i in range(0, 11)]
    variables = ["si10", "dswrf", "t", "prate"]
    ids = np.array(range(0, 10)) + 9905

    coords = (
        ("init_time", init_times),
        ("variable", variables),
        ("step", steps),
        ("id", ids),
    )

    nwp_array_shape = (len(init_times), len(variables), len(steps), len(ids))

    nwp_data = xr.DataArray(
        np.random.uniform(0, 200, size=nwp_array_shape),
        coords=coords,
    )

    nwp_data = nwp_data.to_dataset(name="UKV")
    with tempfile.TemporaryDirectory() as tmpdir:
        filename = tmpdir + "/nwp.zarr"

        nwp_data.to_zarr(filename, engine="h5netcdf")

        yield filename


@pytest.fixture()
def nwp_gfs_data():
    """Create xarray netcdf file for NWP data

    Variables
    - init_time
    - step
    - latitude
    - longitude
    """

    init_times = pd.date_range(start=datetime(2022, 9, 1), freq="60min", periods=24 * 7)
    steps = [timedelta(minutes=60 * i) for i in range(0, 11)]
    x = np.array(range(0, 10))
    y = np.array(range(0, 10))
    variables = ["si10", "dswrf", "t", "prate"]

    coords = (
        ("time", init_times),
        ("step", steps),
        ("longitude", x),
        ("latitude", y),
    )

    nwp_array_shape = (len(init_times), len(steps), len(x), len(y))

    data_arrays = []

    for variable in variables:
        data_arrays += [
            xr.DataArray(
                np.random.uniform(0, 200, size=nwp_array_shape),
                coords=coords,
                name=variable,
            )
        ]

    nwp_data = xr.merge(data_arrays)
    return nwp_data


@pytest.fixture()
def nwp_gfs_data_filename(nwp_gfs_data):
    with tempfile.TemporaryDirectory() as tmpdir:
        filename = tmpdir + "/nwp.zarr"
        nwp_gfs_data.to_zarr(filename)
        yield filename


@pytest.fixture()
def nwp_ukv_data():
    init_times = pd.date_range(start=datetime(2022, 9, 1), freq="180min", periods=24 * 7)
    steps = [timedelta(minutes=60 * i) for i in range(0, 11)]

    # These are the values from the training data but it takes too long:
    # -> x = np.arange(-239_000, 857_000, 2000) # Shape:  (548,)
    # -> y = np.arange(-183_000, 1225_000, 2000)[::-1] # Shape:  (704,)

    # This is much faster:
    x = np.linspace(-239_000, 857_000, 100)
    y = np.linspace(-183_000, 1225_000, 100)[::-1]  # UKV data must run top to bottom
    variables = ["si10", "dswrf", "t", "prate"]

    coords = (
        ("init_time", init_times),
        ("variable", variables),
        ("step", steps),
        ("x", x),
        ("y", y),
    )

    nwp_array_shape = (len(init_times), len(variables), len(steps), len(x), len(y))

    nwp_data = xr.DataArray(
        np.random.uniform(0, 200, size=nwp_array_shape),
        coords=coords,
    )
    return nwp_data.to_dataset(name="UKV")


@pytest.fixture()
def nwp_ukv_data_filename(nwp_ukv_data):
    with tempfile.TemporaryDirectory() as tmpdir:
        filename = tmpdir + "/nwp.zarr"
        nwp_ukv_data.to_zarr(filename)
        yield filename


@pytest.fixture()
def configuration_filename():
    return f"{_top_test_directory}/data/configs/test.yaml"


@pytest.fixture()
def configuration():
    filename = f"{_top_test_directory}/data/configs/test.yaml"
    return load_yaml_configuration(filename=filename)


@pytest.fixture()
def configuration_no_gsp():
    filename = f"{_top_test_directory}/data/configs/wind_test.yaml"
    return load_yaml_configuration(filename=filename)


@pytest.fixture()
def configuration_with_pv_netcdf(pv_netcdf_file):
    filename = f"{_top_test_directory}/data/configs/test.yaml"

    configuration = load_yaml_configuration(filename=filename)
    with tempfile.TemporaryDirectory() as tmpdir:
        configuration_filename = tmpdir + "/configuration.yaml"
        configuration.input_data.pv.pv_files_groups = [
            configuration.input_data.pv.pv_files_groups[0]
        ]
        configuration.input_data.pv.pv_files_groups[0].pv_filename = pv_netcdf_file
        save_yaml_configuration(configuration=configuration, filename=configuration_filename)

        yield configuration_filename


@pytest.fixture()
def configuration_with_pv_netcdf_and_nwp(pv_netcdf_file, nwp_ukv_data_filename):
    filename = f"{_top_test_directory}/data/configs/test.yaml"

    configuration = load_yaml_configuration(filename=filename)
    with tempfile.TemporaryDirectory() as tmpdir:
        configuration_filename = tmpdir + "/configuration.yaml"
        configuration.input_data.pv.pv_files_groups[0].pv_filename = pv_netcdf_file
        configuration.input_data.pv.pv_files_groups = [
            configuration.input_data.pv.pv_files_groups[0]
        ]
        configuration.input_data.nwp["ukv"].nwp_zarr_path = nwp_ukv_data_filename
        save_yaml_configuration(configuration=configuration, filename=configuration_filename)

        yield configuration_filename


@pytest.fixture()
def configuration_with_gsp_and_nwp(gsp_zarr_file, nwp_ukv_data_filename):
    filename = f"{_top_test_directory}/data/configs/test.yaml"

    configuration = load_yaml_configuration(filename=filename)
    with tempfile.TemporaryDirectory() as tmpdir:
        configuration_filename = tmpdir + "/configuration.yaml"
        configuration.input_data.gsp.gsp_zarr_path = gsp_zarr_file
        configuration.input_data.nwp["ukv"].nwp_zarr_path = nwp_ukv_data_filename
        save_yaml_configuration(configuration=configuration, filename=configuration_filename)

        yield configuration_filename


@pytest.fixture()
def wind_configuration_filename():
    return f"{_top_test_directory}/data/configs/wind_test.yaml"



"""Load ASOS data from local files for training/inference"""
import logging

import fsspec
import xarray as xr
from torch.utils.data import IterDataPipe, functional_datapipe

from ocf_datapipes.config.model import Sensor

_log = logging.getLogger(__name__)


@functional_datapipe("OpenAWOS")
class OpenAWOSFromNetCDFIterDataPipe(IterDataPipe):
    """OpenAWOSFromNetCDFIterDataPipe"""

    def __init__(
        self,
        sensor: Sensor,
    ):
        """
        Datapipe to load sensor data from AWOS network

        Args:
            sensor: Sensor configuration
        """
        super().__init__()
        self.sensor = sensor
        self.filename = self.sensor.sensor_filename
        self.variables = list(self.sensor.sensor_variables)

    def __iter__(self):
        with fsspec.open(self.filename, "rb") as f:
            ds = xr.open_dataset(f)
            # Rename timestamp to time_utc
            ds = ds.rename({"timestamp": "time_utc"})
            # Add coordinate to station_id dimension
            ds = ds.assign_coords(station_id=list(range(len(ds.station_id))))
            # Get all indicies where the latitude is NaN
            nan_lat = ds.latitude.isnull().values
            nan_lat_indicies = nan_lat.nonzero()[0]
            ds = ds.drop_isel(station_id=nan_lat_indicies)
            # Only keep wind speed
            ds = ds[self.variables].to_array()
        while True:
            yield ds



"""Wrapper for Power Perceiver Production Data Pipeline"""
import logging
from datetime import timedelta
from pathlib import Path
from typing import Union

import xarray
from torch.utils.data.datapipes.datapipe import IterDataPipe

import ocf_datapipes  # noqa
from ocf_datapipes.batch import MergeNumpyModalities, MergeNWPNumpyModalities
from ocf_datapipes.config.load import load_yaml_configuration
from ocf_datapipes.config.model import Configuration
from ocf_datapipes.load import OpenNWP, OpenPVFromNetCDF
from ocf_datapipes.training.common import normalize_pv
from ocf_datapipes.utils.consts import NWP_MEANS, NWP_STDS

logger = logging.getLogger(__name__)
xarray.set_options(keep_attrs=True)

BUFFER_SIZE = 5


def pv_nwp_datapipe(
    configuration_filename: Union[Path, str],
) -> IterDataPipe:
    """Create datapipe which yields PV and NWP data

    This could be used to predict PV data from NWP inputs

    Args:
        configuration_filename: Name of the configuration

    Returns:
        DataPipe ready to be put in a Dataloader for production
    """

    # Load configuration
    configuration: Configuration = load_yaml_configuration(configuration_filename)

    # Load and normalize PV data, and fill night-time NaNs
    pv_datapipe = (
        OpenPVFromNetCDF(pv=configuration.input_data.pv)
        .pv_fill_night_nans()
        .normalize(normalize_fn=normalize_pv)
    )

    # Load and normalize NWP data - There may be multiple NWP sources
    nwp_datapipe_dict = {}
    for nwp_source, nwp_conf in configuration.input_data.nwp.items():
        nwp_datapipe_dict[nwp_source] = OpenNWP(
            nwp_conf.nwp_zarr_path, provider=nwp_conf.nwp_provider
        ).normalize(mean=NWP_MEANS[nwp_conf.nwp_provider], std=NWP_STDS[nwp_conf.nwp_provider])

    # Add t0 idx to PV
    pv_datapipe = pv_datapipe.add_t0_idx_and_sample_period_duration(
        sample_period_duration=timedelta(
            minutes=configuration.input_data.pv.time_resolution_minutes
        ),
        history_duration=timedelta(minutes=configuration.input_data.pv.history_minutes),
    )

    # Add t0 idx to NWP and fork for time periods overlap
    nwp_time_periods_datapipes = {}
    for nwp_source, nwp_conf in configuration.input_data.nwp.items():
        nwp_datapipe_dict[nwp_source], nwp_time_periods_datapipes[nwp_source] = (
            nwp_datapipe_dict[nwp_source]
            .add_t0_idx_and_sample_period_duration(
                sample_period_duration=timedelta(hours=3),
                history_duration=timedelta(minutes=nwp_conf.history_minutes),
            )
            .fork(2, buffer_size=BUFFER_SIZE)
        )

    # Fork datapipe for raw data, time slicing, and location slicing uses
    (
        pv_datapipe,
        pv_time_periods_datapipe,
        pv_valid_times_datapipe,
        pv_location_datapipe,
    ) = pv_datapipe.fork(4, buffer_size=BUFFER_SIZE)

    # Get locations - Construct datapipe yielding the locations to be used for each sample
    location_datapipe = pv_location_datapipe.pick_locations()

    # Take NWP spatial slice
    for nwp_source, nwp_conf in configuration.input_data.nwp.items():
        loc_dp, location_datapipe = location_datapipe.fork(2, buffer_size=BUFFER_SIZE)

        nwp_datapipe_dict[nwp_source] = nwp_datapipe_dict[nwp_source].select_spatial_slice_pixels(
            loc_dp,
            roi_height_pixels=nwp_conf.nwp_image_size_pixels_height,
            roi_width_pixels=nwp_conf.nwp_image_size_pixels_width,
        )

    # Slice PV systems
    pv_datapipe = pv_datapipe.select_id(location_datapipe=location_datapipe, data_source_name="pv")

    # Get contiguous time periods for each data source
    pv_time_periods_datapipe = pv_time_periods_datapipe.find_contiguous_t0_time_periods(
        sample_period_duration=timedelta(
            minutes=configuration.input_data.pv.time_resolution_minutes
        ),
        history_duration=timedelta(minutes=configuration.input_data.pv.history_minutes),
        forecast_duration=timedelta(minutes=configuration.input_data.pv.forecast_minutes),
    )

    for nwp_source, nwp_conf in configuration.input_data.nwp.items():
        nwp_time_periods_datapipes[nwp_source] = nwp_time_periods_datapipes[
            nwp_source
        ].find_contiguous_t0_time_periods(
            sample_period_duration=timedelta(hours=3),
            history_duration=timedelta(minutes=nwp_conf.history_minutes),
            forecast_duration=timedelta(minutes=nwp_conf.forecast_minutes),
            time_dim="init_time_utc",
        )

    # Find joint overlapping time periods
    overlapping_datapipe = pv_time_periods_datapipe.filter_to_overlapping_time_periods(
        secondary_datapipes=[*nwp_time_periods_datapipes.values()],
    )
    pv_time_periods, nwp_time_periods = overlapping_datapipe.fork(2, buffer_size=BUFFER_SIZE)

    # Filter PVs to times valid for all data sources
    pv_valid_times_datapipe = pv_valid_times_datapipe.filter_time_periods(
        time_periods=overlapping_datapipe
    )

    # Select t0 times
    t0_datapipe = pv_valid_times_datapipe.pick_t0_times()

    # Take NWP time slices and convert to NumpyBatch
    nwp_numpy_modalities = dict()

    for nwp_source, nwp_conf in configuration.input_data.nwp.items():
        nwp_t0_datapipe, t0_datapipe = t0_datapipe.fork(2)

        nwp_numpy_modalities[nwp_source] = (
            nwp_datapipe_dict[nwp_source]
            .select_time_slice_nwp(
                t0_datapipe=nwp_t0_datapipe,
                sample_period_duration=timedelta(hours=3),
                history_duration=timedelta(minutes=nwp_conf.history_minutes),
                forecast_duration=timedelta(minutes=nwp_conf.forecast_minutes),
            )
            .convert_nwp_to_numpy_batch()
        )

    # Combine the NWPs into NumpyBatch
    nwps_numpy_datapipe = MergeNWPNumpyModalities(nwp_numpy_modalities)

    # Take PV time slices and convert to NumpyBatch
    pv_numpy_datapipe = pv_datapipe.select_time_slice(
        t0_datapipe=t0_datapipe,
        history_duration=timedelta(minutes=configuration.input_data.pv.history_minutes),
        forecast_duration=timedelta(minutes=configuration.input_data.pv.forecast_minutes),
        sample_period_duration=timedelta(
            minutes=configuration.input_data.pv.time_resolution_minutes
        ),
    ).convert_pv_to_numpy_batch()

    # Join data pipes together, and add solar coords
    combined_datapipe = MergeNumpyModalities([pv_numpy_datapipe, nwps_numpy_datapipe])

    combined_datapipe = combined_datapipe.add_sun_position(modality_name="pv")

    # Batch the data with batch size 4
    combined_datapipe = combined_datapipe.batch(4).merge_numpy_batch()

    return combined_datapipe



"""DWD ICON Loading"""
import pandas as pd
import xarray as xr

from ocf_datapipes.load.nwp.providers.utils import open_zarr_paths


def open_icon_eu(zarr_path) -> xr.Dataset:
    """
    Opens the ICON data

    ICON EU Data is on a regular lat/lon grid
    It has data on multiple pressure levels, as well as the surface
    Each of the variables is its own data variable

    Args:
        zarr_path: Path to the zarr to open

    Returns:
        Xarray DataArray of the NWP data
    """
    # Open the data
    nwp = open_zarr_paths(zarr_path, time_dim="time")
    nwp = nwp.rename({"time": "init_time_utc"})
    # Sanity checks.
    time = pd.DatetimeIndex(nwp.init_time_utc)
    assert time.is_unique
    assert time.is_monotonic_increasing
    return nwp


def open_icon_global(zarr_path) -> xr.Dataset:
    """
    Opens the ICON data

    ICON Global Data is on an isohedral grid, so the points are not regularly spaced
    It has data on multiple pressure levels, as well as the surface
    Each of the variables is its own data variable

    Args:
        zarr_path: Path to the zarr to open

    Returns:
        Xarray DataArray of the NWP data
    """
    # Open the data
    nwp = open_zarr_paths(zarr_path, time_dim="time")
    nwp = nwp.rename({"time": "init_time_utc"})
    # ICON Global archive script didn't define the values to be
    # associated with lat/lon so fixed here
    nwp.coords["latitude"] = (("values",), nwp.latitude.values)
    nwp.coords["longitude"] = (("values",), nwp.longitude.values)
    # Sanity checks.
    time = pd.DatetimeIndex(nwp.init_time_utc)
    assert time.is_unique
    assert time.is_monotonic_increasing
    return nwp



"""Geospatial functions"""
from datetime import datetime
from numbers import Number
from typing import Union

import numpy as np
import pandas as pd
import pvlib
import pyproj
import xarray as xr

# OSGB is also called "OSGB 1936 / British National Grid -- United
# Kingdom Ordnance Survey".  OSGB is used in many UK electricity
# system maps, and is used by the UK Met Office UKV model.  OSGB is a
# Transverse Mercator projection, using 'easting' and 'northing'
# coordinates which are in meters.  See https://epsg.io/27700
OSGB36 = 27700

# WGS84 is short for "World Geodetic System 1984", used in GPS. Uses
# latitude and longitude.
WGS84 = 4326


_osgb_to_lon_lat = pyproj.Transformer.from_crs(
    crs_from=OSGB36, crs_to=WGS84, always_xy=True
).transform
_lon_lat_to_osgb = pyproj.Transformer.from_crs(
    crs_from=WGS84, crs_to=OSGB36, always_xy=True
).transform
_geod = pyproj.Geod(ellps="WGS84")


def osgb_to_lon_lat(
    x: Union[Number, np.ndarray], y: Union[Number, np.ndarray]
) -> tuple[Union[Number, np.ndarray], Union[Number, np.ndarray]]:
    """Change OSGB coordinates to lon, lat.

    Args:
        x: osgb east-west
        y: osgb north-south
    Return: 2-tuple of longitude (east-west), latitude (north-south)
    """
    return _osgb_to_lon_lat(xx=x, yy=y)


def lon_lat_to_osgb(
    x: Union[Number, np.ndarray],
    y: Union[Number, np.ndarray],
) -> tuple[Union[Number, np.ndarray], Union[Number, np.ndarray]]:
    """Change lon-lat coordinates to OSGB.

    Args:
        x: longitude east-west
        y: latitude north-south

    Return: 2-tuple of OSGB x, y
    """
    return _lon_lat_to_osgb(xx=x, yy=y)


def lon_lat_to_geostationary_area_coords(
    x: Union[Number, np.ndarray],
    y: Union[Number, np.ndarray],
    xr_data: Union[xr.Dataset, xr.DataArray],
) -> tuple[Union[Number, np.ndarray], Union[Number, np.ndarray]]:
    """Loads geostationary area and change from lon-lat to geostationaery coords

    Args:
        x: Longitude east-west
        y: Latitude north-south
        xr_data: xarray object with geostationary area

    Returns:
        Geostationary coords: x, y
    """
    # Only load these if using geostationary projection
    import pyproj
    import pyresample

    try:
        area_definition_yaml = xr_data.attrs["area"]
    except KeyError:
        area_definition_yaml = xr_data.data.attrs["area"]
    geostationary_area_definition = pyresample.area_config.load_area_from_string(
        area_definition_yaml
    )
    geostationary_crs = geostationary_area_definition.crs
    lonlat_to_geostationary = pyproj.Transformer.from_crs(
        crs_from=WGS84,
        crs_to=geostationary_crs,
        always_xy=True,
    ).transform
    return lonlat_to_geostationary(xx=x, yy=y)


def osgb_to_geostationary_area_coords(
    x: Union[Number, np.ndarray],
    y: Union[Number, np.ndarray],
    xr_data: Union[xr.Dataset, xr.DataArray],
) -> tuple[Union[Number, np.ndarray], Union[Number, np.ndarray]]:
    """Loads geostationary area and transformation from OSGB to geostationary coords

    Args:
        x: osgb east-west
        y: osgb north-south
        xr_data: xarray object with geostationary area

    Returns:
        Geostationary coords: x, y
    """
    # Only load these if using geostationary projection
    import pyproj
    import pyresample

    try:
        area_definition_yaml = xr_data.attrs["area"]
    except KeyError:
        area_definition_yaml = xr_data.data.attrs["area"]
    geostationary_area_definition = pyresample.area_config.load_area_from_string(
        area_definition_yaml
    )
    geostationary_crs = geostationary_area_definition.crs
    osgb_to_geostationary = pyproj.Transformer.from_crs(
        crs_from=OSGB36, crs_to=geostationary_crs, always_xy=True
    ).transform
    return osgb_to_geostationary(xx=x, yy=y)


def geostationary_area_coords_to_osgb(
    x: Union[Number, np.ndarray],
    y: Union[Number, np.ndarray],
    xr_data: Union[xr.Dataset, xr.DataArray],
) -> tuple[Union[Number, np.ndarray], Union[Number, np.ndarray]]:
    """Loads geostationary area and change from geostationary coords to OSGB

    Args:
        x: geostationary x coord
        y: geostationary y coord
        xr_data: xarray object with geostationary area

    Returns:
        OSGB x, OSGB y
    """
    # Only load these if using geostationary projection
    import pyproj
    import pyresample

    try:
        area_definition_yaml = xr_data.attrs["area"]
    except KeyError:
        area_definition_yaml = xr_data.data.attrs["area"]
    geostationary_area_definition = pyresample.area_config.load_area_from_string(
        area_definition_yaml
    )
    geostationary_crs = geostationary_area_definition.crs
    geostationary_to_osgb = pyproj.Transformer.from_crs(
        crs_from=geostationary_crs, crs_to=OSGB36, always_xy=True
    ).transform
    return geostationary_to_osgb(xx=x, yy=y)


def geostationary_area_coords_to_lonlat(
    x: Union[Number, np.ndarray],
    y: Union[Number, np.ndarray],
    xr_data: Union[xr.Dataset, xr.DataArray],
) -> tuple[Union[Number, np.ndarray], Union[Number, np.ndarray]]:
    """Loads geostationary area and change from geostationary to lon-lat coords

    Args:
        x: geostationary x coord
        y: geostationary y coord
        xr_data: xarray object with geostationary area

    Returns:
        longitude, latitude
    """
    # Only load these if using geostationary projection
    import pyproj
    import pyresample

    try:
        area_definition_yaml = xr_data.attrs["area"]
    except KeyError:
        area_definition_yaml = xr_data.data.attrs["area"]
    geostationary_area_definition = pyresample.area_config.load_area_from_string(
        area_definition_yaml
    )
    geostationary_crs = geostationary_area_definition.crs
    geostationary_to_lonlat = pyproj.Transformer.from_crs(
        crs_from=geostationary_crs, crs_to=WGS84, always_xy=True
    ).transform
    return geostationary_to_lonlat(xx=x, yy=y)


def calculate_azimuth_and_elevation_angle(
    latitude: float, longitude: float, datestamps: list[datetime]
) -> pd.DataFrame:
    """
    Calculation the azimuth angle, and the elevation angle for several datetamps.

    But for one specific lat/lon location

    More details see:
    https://www.celestis.com/resources/faq/what-are-the-azimuth-and-elevation-of-a-satellite/

    Args:
        latitude: latitude of the pv site
        longitude: longitude of the pv site
        datestamps: list of datestamps to calculate the sun angles. i.e the sun moves from east to
            west in the day.

    Returns: Pandas data frame with the index the same as 'datestamps', with columns of
    "elevation" and "azimuth" that have been calculate.

    """
    # get the solor position
    solpos = pvlib.solarposition.get_solarposition(datestamps, latitude, longitude)

    # extract the information we want
    return solpos[["elevation", "azimuth"]]


def move_lon_lat_by_meters(lon, lat, meters_east, meters_north):
    """
    Move a (lon, lat) by a certain number of meters north and east

    Args:
        lon: longitude
        lat: latitude
        meters_east: number of meters to move east
        meters_north: number of meters to move north

    Returns:
        tuple of lon, lat
    """
    new_lon = _geod.fwd(lons=lon, lats=lat, az=90, dist=meters_east)[0]
    new_lat = _geod.fwd(lons=lon, lats=lat, az=0, dist=meters_north)[1]
    return new_lon, new_lat


def _coord_priority(available_coords):
    if "longitude" in available_coords:
        return "lon_lat", "longitude", "latitude"
    elif "x_geostationary" in available_coords:
        return "geostationary", "x_geostationary", "y_geostationary"
    elif "x_osgb" in available_coords:
        return "osgb", "x_osgb", "y_osgb"
    elif "x" in available_coords:
        return "xy", "x", "y"
    else:
        return None, None, None


def spatial_coord_type(ds: xr.Dataset):
    """Searches the dataset to determine the kind of spatial coordinates present.

    This search has a preference for the dimension coordinates of the xarray object. If none of the
    expected coordinates exist in the dimension coordinates, it then searches the non-dimension
    coordinates. See https://docs.xarray.dev/en/latest/user-guide/data-structures.html#coordinates.

    Args:
        ds: Dataset with spatial coords

    Returns:
        str: The kind of the coordinate system
        x_coord: Name of the x-coordinate
        y_coord: Name of the y-coordinate
    """
    if isinstance(ds, xr.DataArray):
        # Search dimension coords of dataarray
        coords = _coord_priority(ds.xindexes)
    elif isinstance(ds, xr.Dataset):
        # Search dimension coords of all variables in dataset
        coords = _coord_priority(set([v for k in ds.keys() for v in list(ds[k].xindexes)]))
    else:
        raise ValueError(f"Unrecognized input type: {type(ds)}")

    if coords == (None, None, None):
        # If no dimension coords found, search non-dimension coords
        coords = _coord_priority(list(ds.coords))

    return coords



"""Selection datapipes"""
from .apply_pv_dropout import ApplyPVDropoutIterDataPipe as ApplyPVDropout
from .apply_standard_dropout import ApplyDropoutTimeIterDataPipe as ApplyDropoutTime
from .apply_standard_dropout import DrawDropoutTimeIterDataPipe as DrawDropoutTime
from .filter_channels import FilterChannelsIterDataPipe as FilterChannels
from .filter_gsp_ids import FilterGSPIDsIterDataPipe as FilterGSPIDs
from .filter_pv_sys_generating_overnight import (
    FilterPvSysGeneratingOvernightIterDataPipe as FilterPvSysGeneratingOvernight,
)
from .filter_pv_sys_with_only_nan_in_a_day import (
    FilterPVSystemsWithOnlyNanInADayIterDataPipe as FilterPVSystemsWithOnlyNanInADay,
)
from .filter_pv_systems_by_capacity import (
    FilterPVSystemsOnCapacityIterDataPipe as FilterPVSystemsOnCapacity,
)
from .filter_time_periods import FilterTimePeriodsIterDataPipe as FilterTimePeriods
from .filter_times import FilterTimesIterDataPipe as SelectTrainTestTimePeriod
from .filter_to_overlapping_time_periods import (
    FilterToOverlappingTimePeriodsIterDataPipe as FilterToOverlappingTimePeriods,
)
from .find_contiguous_t0_time_periods import (
    FindContiguousT0TimePeriodsIterDataPipe as FindContiguousT0TimePeriods,
)
from .find_contiguous_t0_time_periods import (
    FindContiguousT0TimePeriodsNWPIterDataPipe as FindContiguousT0TimePeriodsNWP,
)
from .pick_locations import PickLocationsIterDataPipe as PickLocations
from .pick_locations_and_t0_times import PickLocationsAndT0sIterDataPipe as PickLocationsAndT0s
from .pick_t0_times import PickT0TimesIterDataPipe as PickT0Times
from .select_id import SelectIDIterDataPipe as SelectID
from .select_non_nan_timestamps import SelectNonNaNTimesIterDataPipe as SelectNonNaNTimes
from .select_spatial_slice import SelectSpatialSliceMetersIterDataPipe as SelectSpatialSliceMeters
from .select_spatial_slice import SelectSpatialSlicePixelsIterDataPipe as SelectSpatialSlicePixels
from .select_time_slice import SelectTimeSliceIterDataPipe as SelectTimeSlice
from .select_time_slice_nwp import (
    SelectTimeSliceNWPIterDataPipe as SelectTimeSliceNWP,
)



"""Datapipe and utils to load PV data from NetCDF for training"""
import io
import logging
from pathlib import Path
from typing import Optional, Union

import fsspec
import numpy as np
import pandas as pd
import xarray as xr
from torch.utils.data import IterDataPipe, functional_datapipe

from ocf_datapipes.config.model import PV
from ocf_datapipes.load.pv.utils import (
    put_pv_data_into_an_xr_dataarray,
)

_log = logging.getLogger(__name__)


@functional_datapipe("open_pv_netcdf")
class OpenPVFromNetCDFIterDataPipe(IterDataPipe):
    """Datapipe to load NetCDF"""

    def __init__(
        self,
        pv: PV,
    ):
        """Datapipe to load PV from NetCDF.

        Args:
            pv: pv configuration
        """
        super().__init__()
        self.pv = pv
        self.pv_power_filenames = [
            pv_files_group.pv_filename for pv_files_group in pv.pv_files_groups
        ]
        self.pv_metadata_filenames = [
            pv_files_group.pv_metadata_filename for pv_files_group in pv.pv_files_groups
        ]
        self.inferred_metadata_filenames = [
            pv_files_group.inferred_metadata_filename for pv_files_group in pv.pv_files_groups
        ]
        self.labels = [pv_files_group.label for pv_files_group in pv.pv_files_groups]

    def __iter__(self):
        pv_array_list = []
        for i in range(len(self.pv_power_filenames)):
            pv_array: xr.DataArray = load_everything_into_ram(
                self.pv_power_filenames[i],
                self.pv_metadata_filenames[i],
                inferred_metadata_filename=self.inferred_metadata_filenames[i],
                label=self.labels[i],
            )
            pv_array_list.append(pv_array)

        pv_array = xr.concat(pv_array_list, dim="pv_system_id")

        while True:
            yield pv_array


def load_everything_into_ram(
    generation_filename,
    metadata_filename,
    inferred_metadata_filename: Optional[Union[str, Path]] = None,
    estimated_capacity_percentile: float = 100,
    label: Optional[str] = None,
) -> xr.DataArray:
    """Load PV data into xarray DataArray in RAM.

    Args:
        generation_filename: Filepath to the PV generation data
        metadata_filename: Filepath to the PV metadata
        inferred_metadata_filename: Filepath to inferred metadata
        estimated_capacity_percentile: Percentile used as the estimated capacity for each PV
            system. Recommended range is 99-100.
        label: Label of which provider the PV data came from
    """

    # load metadata
    df_metadata = _load_pv_metadata(metadata_filename, inferred_metadata_filename, label)

    # Load pd.DataFrame of power and pd.Series of capacities:
    df_gen, estimated_capacities = _load_pv_generation_and_capacity(
        generation_filename,
        estimated_capacity_percentile=estimated_capacity_percentile,
    )

    # Drop systems where all values are NaN
    df_gen.dropna(axis="columns", how="all", inplace=True)
    estimated_capacities = estimated_capacities[df_gen.columns]

    # Ensure systems are consistant between generation data, and metadata
    common_systems = list(np.intersect1d(df_metadata.index, df_gen.columns))
    df_gen = df_gen[common_systems]
    df_metadata = df_metadata.loc[common_systems]
    estimated_capacities = estimated_capacities.loc[common_systems]

    # Compile data into an xarray DataArray
    xr_array = put_pv_data_into_an_xr_dataarray(
        df_gen=df_gen,
        observed_system_capacities=estimated_capacities,
        nominal_system_capacities=df_metadata.capacity_watts,
        ml_id=df_metadata.ml_id,
        latitude=df_metadata.latitude,
        longitude=df_metadata.longitude,
        tilt=df_metadata.get("tilt"),
        orientation=df_metadata.get("orientation"),
    )

    # Sanity checks
    time_utc = pd.DatetimeIndex(xr_array.time_utc)
    assert time_utc.is_monotonic_increasing
    assert time_utc.is_unique

    return xr_array


def _load_pv_generation_and_capacity(
    filename: Union[str, Path],
    estimated_capacity_percentile: float = 99,
    label: Optional[str] = None,
) -> tuple[pd.DataFrame, pd.Series]:
    """Load the PV data and estimates the capacity for each PV system.

    The capacity is estimated by taking the max value across all datetimes in the input file.

    Args:
        filename: The filename (netcdf) of the PV data to load
        estimated_capacity_percentile: Percentile used as the estimated capacity for each PV
            system. Recommended range is 99-100.
        label: Label of which provider the PV data came from

    Returns:
        DataFrame of power output in watts. Columns are PV systems, rows are datetimes
        Series of PV system estimated capacities in watts
    """

    _log.info(f"Loading solar PV power data from {filename}.")

    # Load data in a way that will work in the cloud and locally:
    if ".parquet" in str(filename):
        _log.debug(f"Loading PV parquet file {filename}")

        df_raw = pd.read_parquet(filename, engine="fastparquet")

        # Assuming the data is reported 5-minutely
        # The power output in watts is 12 times (60 minutes /5 minutes) the energy in watt-hours
        df_raw["generation_w"] = df_raw["generation_wh"] * 12

        # pivot on ss_id
        df_gen = df_raw.pivot(index="timestamp", columns="ss_id", values="generation_w")
        del df_raw

    else:
        with fsspec.open(filename, mode="rb") as file:
            file_bytes = file.read()

        _log.info("Loaded solar PV power bytes, now converting to xarray")
        with io.BytesIO(file_bytes) as file:
            ds_gen = xr.load_dataset(file, engine="h5netcdf")

        df_gen = ds_gen.to_dataframe()

        if label == "pvoutput.org":
            _log.warning("Converting timezone. ARE YOU SURE THAT'S WHAT YOU WANT TO DO?")
            try:
                df_gen = df_gen.tz_localize("Europe/London").tz_convert("UTC").tz_convert(None)
            except Exception as e:
                _log.warning(
                    "Could not convert timezone from London to UTC. "
                    "Going to try and carry on anyway"
                )
                _log.warning(e)

    # Fix data types
    df_gen = df_gen.astype(np.float32)
    df_gen.columns = df_gen.columns.astype(np.int64)

    _log.info("Loaded solar PV power data and converting to pandas.")
    estimated_capacities = df_gen.quantile(estimated_capacity_percentile / 100)

    # Remove systems with no generation data
    mask = estimated_capacities > 0
    estimated_capacities = estimated_capacities[mask]
    df_gen = df_gen.loc[:, mask]

    _log.info(
        "After loading:"
        f" {len(df_gen)} PV power datetimes."
        f" {len(df_gen.columns)} PV power PV system IDs."
    )

    # Sanity checks
    assert not df_gen.columns.duplicated().any()
    assert not df_gen.index.duplicated().any()
    assert np.isfinite(estimated_capacities).all()
    assert (estimated_capacities > 0).all()
    assert np.array_equal(df_gen.columns, estimated_capacities.index)

    return df_gen, estimated_capacities


# Adapted from nowcasting_dataset.data_sources.pv.pv_data_source
def _load_pv_metadata(
    filename: str, inferred_filename: Optional[str] = None, label: Optional[str] = None
) -> pd.DataFrame:
    """Return pd.DataFrame of PV metadata.

    Shape of the returned pd.DataFrame for Passiv PV data:
        Index: ss_id (Sheffield Solar ID)
        Columns: llsoacd, orientation, tilt, kwp, operational_at, latitude, longitude, system_id,
            ml_id, capacity_watts
    """
    _log.info(f"Loading PV metadata from {filename}")

    index_col = "ss_id" if label == "solar_sheffield_passiv" else "system_id"
    df_metadata = pd.read_csv(filename, index_col=index_col)

    # Drop if exists
    df_metadata.drop(columns="Unnamed: 0", inplace=True, errors="ignore")

    # Add ml_id column if not in metadata already
    if "ml_id" not in df_metadata.columns:
        df_metadata["ml_id"] = np.nan

    if label == "solar_sheffield_passiv":
        # Add capacity in watts
        df_metadata["capacity_watts"] = df_metadata.kwp * 1000
        # Maybe load inferred metadata if passiv
        if inferred_filename is not None:
            df_metadata = _load_inferred_metadata(filename, df_metadata)
    elif label == "india":  # noqa: E721
        # Add capacity in watts
        df_metadata["capacity_watts"] = df_metadata.capacity_watts
    elif label == "pvoutput.org":
        # For PVOutput.org data
        df_metadata["capacity_watts"] = df_metadata.system_size_watts
        # Rename PVOutput.org tilt name to be simpler
        # There is a second degree tilt, but this should be fine for now
        if "array_tilt_degrees" in df_metadata.columns:
            df_metadata["tilt"] = df_metadata["array_tilt_degrees"]

        # Need to change orientation to a number if a string (i.e. SE) that PVOutput.org uses by
        # default
        mapping = {
            "N": 0.0,
            "NE": 45.0,
            "E": 90.0,
            "SE": 135.0,
            "S": 180.0,
            "SW": 225.0,
            "W": 270.0,
            "NW": 315.0,
        }

        # Any other keys other than those in the dict above are mapped to NaN
        df_metadata["orientation"] = df_metadata.orientation.map(mapping)
    else:
        raise NotImplementedError(f"Provider label {label} not implemented")

    _log.info(f"Found {len(df_metadata)} PV systems in {filename}")

    return df_metadata


def _load_inferred_metadata(filename: str, df_metadata: pd.DataFrame) -> pd.DataFrame:
    inferred_metadata = pd.read_csv(filename, index_col="ss_id")
    inferred_metadata = inferred_metadata.rename({"capacity": "kwp"})
    # Replace columns with new data if in the PV metadata already
    df_metadata.update(inferred_metadata)
    return df_metadata



"""
Drop PV systems which has only NaN's in a single day
"""
import logging

import numpy as np
import xarray as xr
from torch.utils.data import IterDataPipe, functional_datapipe

logger = logging.getLogger(__name__)


def return_system_indices_which_has_contiguous_nan(
    arr: np.ndarray, check_interval: int = 287
) -> np.ndarray:
    """This function return system indices

    Returns indexes of system id's in which if they have
    contigous 289 NaN's.

    Args:
        arr: Array of each system pvoutput values for a single day
        check_interval: time range intervals respectively

    """
    # Checking the shape of the input array
    # The array would be a 2d-array which consists of number of (time_utc, pv_system_id)
    number_of_systems = arr.shape[1]

    system_index_values_to_be_dropped = []
    for i in range(0, number_of_systems):
        # For each system id
        single_system_single_day_pv_values = arr[:, i]

        # This loop checks NaN in every element in the array and if the count of NaN
        # is equal to defined interval, it stores the index of the pv system
        mask = np.concatenate(([False], np.isnan(single_system_single_day_pv_values), [False]))
        if ~mask.any():
            continue
        else:
            idx = np.nonzero(mask[1:] != mask[:-1])[0]
            max_count = (idx[1::2] - idx[::2]).max()

        if max_count == check_interval:
            system_index_values_to_be_dropped.append(i)

    return system_index_values_to_be_dropped


@functional_datapipe("filter_pv_systems_with_only_nan_in_a_day")
class FilterPVSystemsWithOnlyNanInADayIterDataPipe(IterDataPipe):
    """Remove any PV systems with less than 1 day of data"""

    def __init__(self, source_datapipe: IterDataPipe, minimum_number_data_points: int) -> None:
        """Remove any PV systems with less than 1 day of data

        This is done, by counting all the non values and check the
        count is greater than 289 (number of 5 minute intervals in a day)

        Args:
            source_datapipe: Datapipe of Xarray Dataset emitting timeseries data
            minimum_number_data_points: Minimum number of intervals in a given day
                For 5min time series, intervals = 288
                For 30min time series, intervals = 48
        """

        self.source_datapipe = source_datapipe
        self.intervals = minimum_number_data_points

    def __iter__(self) -> xr.Dataset():
        # Reading the DataArray
        for xr_dataset in self.source_datapipe:
            dates_array = xr_dataset.coords["time_utc"].values

            # Checking length of time series longer than standard intervals
            if dates_array.size >= self.intervals:
                # Collecting the time series data from 'time_utc' coordinate
                for i in range(0, dates_array.size, self.intervals):
                    if i == self.intervals:
                        break
                    else:
                        # Getting the first and last timestamp of the day based on intervals
                        # Slicing the dataset by individual first timestamp in a day
                        # and last timestamp of the day
                        xr_ds = xr_dataset.sel(
                            time_utc=slice(dates_array[i], dates_array[i + (self.intervals - 1)])
                        )

                        # Extracting system ids with continous NaN's
                        sys_ids = return_system_indices_which_has_contiguous_nan(
                            xr_ds.values, check_interval=self.intervals
                        )

                        # Dropping the systems which are inactive for the whole day
                        if not len(sys_ids) == 0:
                            xr_dataset = xr_dataset.drop_isel(pv_system_id=sys_ids)
            else:
                pass

            yield xr_dataset



""" Make fake NWP data """
from datetime import datetime

import numpy as np

from ocf_datapipes.batch import BatchKey, NWPBatchKey
from ocf_datapipes.batch.fake.utils import get_n_time_steps_from_config, make_time_utc
from ocf_datapipes.config.model import Configuration


def make_fake_nwp_data(
    configuration: Configuration, t0_datetime_utc: datetime, batch_size: int = 8
) -> dict:
    """
    Make Fake NWP data ready for ML model

    Args:
        configuration: configuration object
        t0_datetime_utc: one datetime for when t0 is
        batch_size: Integer batch size to create

    Returns: dictionary of nwp items
    """

    nwp_config = configuration.input_data.nwp

    if configuration.input_data.nwp is None:
        return {}

    batch = {}

    for nwp_source, nwp_config in configuration.input_data.nwp.items():
        source_batch = {}

        n_channels = len(nwp_config.nwp_channels)
        n_y_osgb = nwp_config.nwp_image_size_pixels_height
        n_x_osgb = nwp_config.nwp_image_size_pixels_width
        n_fourier_features = 8

        # make time matrix
        time_utc = make_time_utc(
            batch_size=batch_size,
            history_minutes=nwp_config.history_minutes,
            forecast_minutes=nwp_config.forecast_minutes,
            t0_datetime_utc=t0_datetime_utc,
            time_resolution_minutes=nwp_config.time_resolution_minutes,
        )
        n_times = time_utc.shape[1]

        # main nwp components

        source_batch[
            NWPBatchKey.nwp_init_time_utc
        ] = time_utc  # Seconds since UNIX epoch (1970-01-01).
        source_batch[
            NWPBatchKey.nwp_target_time_utc
        ] = time_utc  # Seconds since UNIX epoch (1970-01-01).
        source_batch[NWPBatchKey.nwp] = np.random.random(
            (batch_size, n_times, n_channels, n_y_osgb, n_x_osgb)
        )
        source_batch[NWPBatchKey.nwp_t0_idx] = get_n_time_steps_from_config(
            input_data_configuration=nwp_config, include_forecast=False
        )

        source_batch[NWPBatchKey.nwp_step] = np.random.randint(0, 100, (batch_size, n_times))
        source_batch[NWPBatchKey.nwp_y_osgb] = np.random.randint(0, 100, (batch_size, n_y_osgb))
        source_batch[NWPBatchKey.nwp_x_osgb] = np.random.randint(0, 100, (batch_size, n_x_osgb))
        source_batch[NWPBatchKey.nwp_channel_names] = np.random.randint(0, 100, (n_channels,))

        # fourier components
        source_batch[NWPBatchKey.nwp_x_osgb_fourier] = np.random.random(
            (batch_size, n_x_osgb, n_fourier_features)
        )
        source_batch[NWPBatchKey.nwp_y_osgb_fourier] = np.random.random(
            (batch_size, n_y_osgb, n_fourier_features)
        )
        source_batch[NWPBatchKey.nwp_target_time_utc] = np.random.random(
            (batch_size, n_times, n_fourier_features)
        )
        source_batch[NWPBatchKey.nwp_init_time_utc] = np.random.random(
            (batch_size, n_times, n_fourier_features)
        )

        batch[nwp_source] = source_batch

    return {BatchKey.nwp: batch}



"""Datapipes for batching together data"""
from .batches import BatchKey, NumpyBatch, NWPBatchKey, NWPNumpyBatch, XarrayBatch
from .merge_numpy_examples_to_batch import (
    MergeNumpyBatchIterDataPipe as MergeNumpyBatch,
)
from .merge_numpy_examples_to_batch import (
    MergeNumpyExamplesToBatchIterDataPipe as MergeNumpyExamplesToBatch,
)
from .merge_numpy_examples_to_batch import (
    stack_np_examples_into_batch,
    unstack_np_batch_into_examples,
)
from .merge_numpy_modalities import MergeNumpyModalitiesIterDataPipe as MergeNumpyModalities
from .merge_numpy_modalities import MergeNWPNumpyModalitiesIterDataPipe as MergeNWPNumpyModalities



"""Converts pressure levels to separate variables"""
from typing import Optional

import xarray as xr
from torch.utils.data import IterDataPipe, functional_datapipe


@functional_datapipe("convert_pressure_levels_to_separate_variables")
class ConvertPressureLevelsToSeparateVariablesIterDataPipe(IterDataPipe):
    """Converts pressure levels to separate variables"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        pressure_level_name: str = "isobaricInhPa",
        pressure_level_to_use: Optional[list] = None,
    ):
        """
        Converts pressure levels to separate variables

        Args:
            source_datapipe: Datapipe of NWP data with pressure levels
            pressure_level_name: Name of the pressure level variable
            pressure_level_to_use: List of pressure levels to use
        """
        super().__init__()
        self.source_datapipe = source_datapipe
        self.pressure_level_to_use = pressure_level_to_use
        self.pressure_level_name = pressure_level_name
        # For icosohedral grids, the lat/lon points are in one large array, not separate

    def __iter__(self) -> xr.DataArray:
        for xr_dataset in self.source_datapipe:
            # Select the given pressure levels
            if self.pressure_level_to_use is not None:
                xr_dataset = xr_dataset.sel({self.pressure_level_name: self.pressure_level_to_use})
            # Unstack the pressure levels into separate variables
            xr_dataarray = xr_dataset.to_stacked_array(
                "level",
                sample_dims=[v for v in xr_dataset.dims if v != self.pressure_level_name],
            )
            # Note: NaN values are surface variables
            yield xr_dataarray



"""Merge multiple modalities together in NumpyBatch"""
from torch.utils.data import IterDataPipe, functional_datapipe

from ocf_datapipes.batch import BatchKey, NumpyBatch
from ocf_datapipes.utils import Zipper


@functional_datapipe("merge_numpy_modalities")
class MergeNumpyModalitiesIterDataPipe(IterDataPipe):
    """Merge multiple modalities together in NumpyBatch"""

    def __init__(self, source_datapipes: [IterDataPipe]):
        """
        Merge multiple modalities together in NumpyBatch

        Args:
            source_datapipes: Set of datapipes to merge emitting NumpyBatch objects
        """
        self.source_datapipes = source_datapipes

    def __iter__(self) -> NumpyBatch:
        """Merge multiple modalities together in NumpyBatch"""
        for np_batches in Zipper(*self.source_datapipes):
            example: NumpyBatch = {}
            for np_batch in np_batches:
                example.update(np_batch)
            yield example


@functional_datapipe("merge_nwp_numpy_modalities")
class MergeNWPNumpyModalitiesIterDataPipe(IterDataPipe):
    """Merge multiple NWPNumpyBatches into a NumpyBatch"""

    def __init__(self, datapipes_dict: dict[IterDataPipe]):
        """
        Merge multiple NWPNumpyBatches into a NumpyBatch

        Args:
            datapipes_dict: dict of datapipes to merge emitting NWPNumpyBatch objects
        """
        self.datapipes_dict = datapipes_dict

    def __iter__(self) -> NumpyBatch:
        """Merge multiple NWPNumpyBatches into a NumpyBatch"""
        keys = list(self.datapipes_dict.keys())
        datapipes = [self.datapipes_dict[k] for k in keys]
        for nwp_np_batches in Zipper(*datapipes):
            example: NumpyBatch = {BatchKey.nwp: {k: v for k, v in zip(keys, nwp_np_batches)}}
            yield example



"""Stacking Xarray objects to Numpy inputs"""
from typing import List

import numpy as np
from torch.utils.data import IterDataPipe, functional_datapipe

from ocf_datapipes.utils import Zipper


@functional_datapipe("stack_xarray")
class StackXarrayIterDataPipe(IterDataPipe):
    """Stack Xarray datasets together"""

    def __init__(
        self,
        source_datapipes: List[IterDataPipe],
    ):
        """

        Processes set of Xarray datasets into Numpy

        Args:
            source_datapipes: Datapipes that emit xarray datasets
                with latitude/longitude coordinates included
        """
        self.source_datapipes = source_datapipes

    def __iter__(self) -> np.ndarray:
        for xr_datas in Zipper(*self.source_datapipes):
            stack = []
            for xr_index, xr_data in enumerate(xr_datas):
                # Resamples to the same number of pixels for both center and contexts
                xr_data = xr_data.to_numpy()
                if len(xr_data.shape) == 2:  # Need to add channel dimension
                    xr_data = np.expand_dims(xr_data, axis=0)
                if len(xr_data.shape) == 3:  # Need to add channel dimension
                    xr_data = np.expand_dims(xr_data, axis=1)
                stack.append(xr_data)
            # Pad out time dimension to be the same, using the largest one
            # All should have 4 dimensions at this point
            max_time_len = np.max([c.shape[0] for c in stack])
            for i in range(len(stack)):
                stack[i] = np.pad(
                    stack[i],
                    pad_width=(
                        (0, max_time_len - stack[i].shape[0]),
                        (0, 0),
                        (0, 0),
                        (0, 0),
                    ),
                    mode="constant",
                    constant_values=0.0,
                )

            stacked_data = np.concatenate([*stack], axis=1)
            yield stacked_data



""" Utils for GSP loading"""
import geopandas as gpd
import numpy as np
import pandas as pd
import xarray as xr


def put_gsp_data_into_an_xr_dataarray(
    gsp_pv_power_mw: np.ndarray,
    time_utc: np.ndarray,
    gsp_id: np.ndarray,
    x_osgb: np.ndarray,
    y_osgb: np.ndarray,
    nominal_capacity_mwp: np.ndarray,
    effective_capacity_mwp: np.ndarray,
) -> xr.DataArray:
    """
    Converts the GSP data to Xarray DataArray

    Args:
        gsp_pv_power_mw: GSP PV Power
        time_utc: Time in UTC
        gsp_id: Id of the GSPs
        x_osgb: OSGB X coordinates
        y_osgb: OSGB y coordinates
        nominal_capacity_mwp: Installed capacity of each GSP
        effective_capacity_mwp: Estimated effective capacity of each GSP

    Returns:
        Xarray DataArray of the GSP data
    """
    # Convert to xr.DataArray:
    data_array = xr.DataArray(
        gsp_pv_power_mw,
        coords=(("time_utc", time_utc), ("gsp_id", gsp_id)),
        name="gsp_pv_power_mw",
    )
    data_array = data_array.assign_coords(
        x_osgb=("gsp_id", x_osgb),
        y_osgb=("gsp_id", y_osgb),
        nominal_capacity_mwp=(("time_utc", "gsp_id"), nominal_capacity_mwp),
        effective_capacity_mwp=(
            ("time_utc", "gsp_id"),
            effective_capacity_mwp,
        ),
    )
    return data_array


def get_gsp_id_to_shape(
    gsp_id_to_region_id_filename: str, sheffield_solar_region_path: str
) -> gpd.GeoDataFrame:
    """
    Get the GSP ID to the shape

    Args:
        gsp_id_to_region_id_filename: Filename of the mapping file
        sheffield_solar_region_path: Path to the region shapes

    Returns:
        GeoDataFrame containing the mapping from ID to shape
    """
    # Load mapping from GSP ID to Sheffield Solar GSP ID to GSP name:
    gsp_id_to_region_id = pd.read_csv(
        gsp_id_to_region_id_filename,
        usecols=["gsp_id", "gsp_name"],
        dtype={"gsp_id": np.int64, "gsp_name": str},
    )

    # Load Sheffield Solar region shapes (which are already in OSGB36 CRS).
    ss_regions = gpd.read_file(sheffield_solar_region_path)

    # Some GSPs are represented by multiple shapes. To find the correct centroid,
    # we need to find the spatial union of those regions, and then find the centroid
    # of those spatial unions. `dissolve(by="GSPs")` groups by "GSPs" and gets
    # the spatial union.
    ss_regions = ss_regions.dissolve(by="GSPs")

    # Merge, so we have a mapping from GSP ID to SS region shape:
    gsp_id_to_shape = (
        ss_regions.merge(gsp_id_to_region_id, left_on="GSPs", right_on="gsp_name")
        .set_index("gsp_id")[["geometry"]]
        .sort_index()
    )

    gsp_0 = (
        gpd.GeoDataFrame(
            data={"gsp_id": [0], "geometry": [gsp_id_to_shape["geometry"].unary_union]}
        )
        .set_index("gsp_id")
        .set_crs(gsp_id_to_shape.crs)
    )

    # For the national forecast, GSP ID 0, we want the shape to be the
    # union of all the other shapes
    gsp_id_to_shape = pd.concat([gsp_id_to_shape, gsp_0]).sort_index()
    return gsp_id_to_shape



"""
This is a class function that drops the pv systems with generates power over night.
"""
import logging

import xarray as xr
from torch.utils.data import IterDataPipe, functional_datapipe

logger = logging.getLogger(__name__)


@functional_datapipe("filter_night_pv")
class FilterPvSysGeneratingOvernightIterDataPipe(IterDataPipe):
    """
    Drop the pv systems which generates power over night date from a timeseries xarray Dataset.
    """

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        threshold=0.2,
        daynight_method: str = "simple",
    ):
        """Drops the PV systems producing output overnight.

        Args:
            source_datapipe: A datapipe that emits xarray Dataset of PV generation
            threshold: Relative threshold for night-time production. Any system that generates more
                power than this in any night-time timestamp will be dropped
            daynight_method: Method used to assign datetimes to either 'night' or 'day'. Either
                "simple" or "elevation". See `AssignDayNightStatusIterDataPipe` for details
        """
        assert daynight_method in ["simple", "elevation"]
        self.source_datapipe = source_datapipe
        self.threshold = threshold
        self.daynight_method = daynight_method

    def __iter__(self) -> xr.DataArray():
        # TODO: Make more general

        for ds in self.source_datapipe.assign_daynight_status(self.daynight_method):
            # Select the night-time values
            ds_night = ds.where(ds.status_daynight == "night", drop=True)

            # Find relative maximum night-time generation for each system
            night_time_max_gen = (ds_night / ds_night.observed_capacity_wp).max(dim="time_utc")

            # Find systems above threshold
            mask = night_time_max_gen > self.threshold

            logger.info(
                f"Dropping {mask.sum().item()} PV systems with IDs:"
                f"{mask.where(mask, drop=True).pv_system_id.values}"
            )
            ds = ds.where(~mask, drop=True)
            yield ds



"""Convert NWP to NumpyBatch"""
import numpy as np
from torch.utils.data import IterDataPipe, functional_datapipe

from ocf_datapipes.batch import NWPBatchKey, NWPNumpyBatch
from ocf_datapipes.utils.utils import datetime64_to_float


@functional_datapipe("convert_nwp_to_numpy_batch")
class ConvertNWPToNumpyBatchIterDataPipe(IterDataPipe):
    """Convert NWP Xarray objects to NWPNumpyBatch"""

    def __init__(self, source_datapipe: IterDataPipe):
        """
        Convert NWP Xarray objecs to NWPNumpyBatch ones

        Args:
            source_datapipe: Datapipe emitting NWP Xarray objects
        """
        super().__init__()
        self.source_datapipe = source_datapipe

    def __iter__(self) -> NWPNumpyBatch:
        """Convert from Xarray to NWPBatchKey"""
        for xr_data in self.source_datapipe:
            example: NWPNumpyBatch = {
                NWPBatchKey.nwp: xr_data.values,
                NWPBatchKey.nwp_t0_idx: xr_data.attrs["t0_idx"],
            }
            if "target_time_utc" in xr_data.coords:
                target_time = xr_data.target_time_utc.values
                example[NWPBatchKey.nwp_target_time_utc] = datetime64_to_float(target_time)
            example[NWPBatchKey.nwp_channel_names] = xr_data.channel.values
            example[NWPBatchKey.nwp_step] = (xr_data.step.values / np.timedelta64(1, "h")).astype(
                np.int64
            )
            example[NWPBatchKey.nwp_init_time_utc] = datetime64_to_float(
                xr_data.init_time_utc.values
            )

            for batch_key, dataset_key in (
                (NWPBatchKey.nwp_y_osgb, "y_osgb"),
                (NWPBatchKey.nwp_x_osgb, "x_osgb"),
            ):
                if dataset_key in xr_data.coords.keys():
                    example[batch_key] = xr_data[dataset_key].values

            yield example



""" Loading configuration functions """
import logging
from typing import Union

import fsspec
from pathy import Pathy
from pyaml_env import parse_config

from ocf_datapipes.config.model import Configuration

logger = logging.getLogger(__name__)


def load_yaml_configuration(filename: Union[str, Pathy]) -> Configuration:
    """
    Load a yaml file which has a configuration in it

    Args:
        filename: the file name that you want to load.  Will load from local, AWS, or GCP
            depending on the protocol suffix (e.g. 's3://bucket/config.yaml').

    Returns:pydantic class

    """
    # load the file to a dictionary
    with fsspec.open(filename, mode="r") as stream:
        configuration = parse_config(data=stream)
        # this means we can load ENVs in the yaml file
    # turn into pydantic class
    configuration = Configuration(**configuration)
    return configuration



"""Common NWP providers"""
import xarray as xr


def open_zarr_paths(zarr_path, time_dim="init_time") -> xr.Dataset:
    """
    Opens the NWP data

    Args:
        zarr_path: Path to the zarr(s) to open
        time_dim: Name of the time dimension

    Returns:
        The opened Xarray Dataset
    """
    if type(zarr_path) in [list, tuple] or "*" in str(zarr_path):  # Multi-file dataset
        nwp = xr.open_mfdataset(
            zarr_path,
            engine="zarr",
            concat_dim=time_dim,
            combine="nested",
            chunks="auto",
        ).sortby(time_dim)
    else:
        nwp = xr.open_dataset(
            zarr_path,
            engine="zarr",
            consolidated=True,
            mode="r",
            chunks="auto",
        )
    return nwp



""" Make fake Sun data """
import numpy as np

from ocf_datapipes.batch import BatchKey, NWPBatchKey
from ocf_datapipes.batch.fake.utils import get_n_time_steps_from_config
from ocf_datapipes.config.model import Configuration


def make_fake_sun_data(configuration: Configuration, batch_size: int = 8):
    """
    Make Fake Sun data ready for ML model. This makes data across all different data inputs

    Args:
        configuration: configuration object
        t0_datetime_utc: one datetime for when t0 is
        batch_size: Integer batch size to create

    Returns: dictionary of pv items
    """

    batch = {}

    # HRV Satellite
    if configuration.input_data.hrvsatellite is not None:
        n_hrv_satellite_timesteps = get_n_time_steps_from_config(
            configuration.input_data.hrvsatellite
        )
        batch[BatchKey.hrvsatellite_solar_azimuth] = np.random.random(
            (batch_size, n_hrv_satellite_timesteps)
        )
        batch[BatchKey.hrvsatellite_solar_elevation] = np.random.random(
            (batch_size, n_hrv_satellite_timesteps)
        )

    # Satellite
    if configuration.input_data.satellite is not None:
        n_satellite_timesteps = get_n_time_steps_from_config(configuration.input_data.satellite)
        batch[BatchKey.satellite_solar_azimuth] = np.random.random(
            (batch_size, n_satellite_timesteps)
        )
        batch[BatchKey.satellite_solar_elevation] = np.random.random(
            (batch_size, n_satellite_timesteps)
        )

    # GSP
    if configuration.input_data.gsp is not None:
        n_satellite_timesteps = get_n_time_steps_from_config(configuration.input_data.gsp)
        batch[BatchKey.gsp_solar_azimuth] = np.random.random((batch_size, n_satellite_timesteps))
        batch[BatchKey.gsp_solar_elevation] = np.random.random((batch_size, n_satellite_timesteps))

    # PV
    if configuration.input_data.pv is not None:
        n_pv_timesteps = get_n_time_steps_from_config(configuration.input_data.pv)
        batch[BatchKey.pv_solar_azimuth] = np.random.random((batch_size, n_pv_timesteps))
        batch[BatchKey.pv_solar_elevation] = np.random.random((batch_size, n_pv_timesteps))

    # NWP
    if configuration.input_data.nwp is not None:
        batch[BatchKey.nwp] = {}

        for nwp_source, nwp_config in configuration.input_data.nwp.items():
            batch[BatchKey.nwp][nwp_source] = {}

            n_nwp_timesteps = get_n_time_steps_from_config(configuration.input_data.nwp[nwp_source])
            batch[BatchKey.nwp][nwp_source][
                NWPBatchKey.nwp_target_time_solar_azimuth
            ] = np.random.random((batch_size, n_nwp_timesteps))
            batch[BatchKey.nwp][nwp_source][
                NWPBatchKey.nwp_target_time_solar_elevation
            ] = np.random.random((batch_size, n_nwp_timesteps))

    return batch



""" Utils Functions to for fake data """
from datetime import timedelta

import numpy as np
import pandas as pd


def make_t0_datetimes_utc(batch_size, temporally_align_examples: bool = False):
    """
    Make list of t0 datetimes

    Args:
        batch_size: the batch size
        temporally_align_examples: option to align examples (within the batch) in time

    Returns: pandas index of t0 datetimes
    """

    all_datetimes = pd.date_range("2023-01-01", "2023-02-01", freq="5min")

    if temporally_align_examples:
        t0_datetimes_utc = list(np.random.choice(all_datetimes, size=1)) * batch_size
    else:
        if len(all_datetimes) >= batch_size:
            replace = False
        else:
            # there are not enought data points,
            # so some examples will have the same datetime
            replace = True

        t0_datetimes_utc = np.random.choice(all_datetimes, batch_size, replace=replace)
    # np.random.choice turns the pd.Timestamp objects into datetime.datetime objects.

    t0_datetimes_utc = pd.to_datetime(t0_datetimes_utc)

    # TODO make test repeatable using numpy generator
    # https://github.com/openclimatefix/nowcasting_dataset/issues/594

    return t0_datetimes_utc


def get_n_time_steps_from_config(input_data_configuration, include_forecast: bool = True) -> int:
    """
    Get the number of time steps from the input data configuration

    Args:
        input_data_configuration: NWP, GSP, PV, e.t.c see ocf_datapipes.config.model
        include_forecast: option to include forecast timesteps or not

    Returns: number of time steps for this input data

    """
    # get history time steps
    n_time_steps = int(
        input_data_configuration.history_minutes / input_data_configuration.time_resolution_minutes
    )

    if include_forecast:
        n_time_steps = n_time_steps + int(
            input_data_configuration.forecast_minutes
            / input_data_configuration.time_resolution_minutes
        )

    # add extra step for now
    n_time_steps = n_time_steps + 1

    return n_time_steps


def make_time_utc(
    batch_size, history_minutes, forecast_minutes, t0_datetime_utc, time_resolution_minutes
):
    """
    Make time utc array

    Args:
        batch_size: the batch size
        history_minutes: the history minutes we want
        forecast_minutes: the amount of minutes that the forecast will be fore
        t0_datetime_utc: t0_datetime
        time_resolution_minutes: time resolution e.g 5 for daat in 5 minutes chunks

    Returns: array of time_utc

    """
    start_datetime = t0_datetime_utc - timedelta(minutes=history_minutes)
    end_datetime = t0_datetime_utc + timedelta(minutes=forecast_minutes)
    time_utc = pd.date_range(
        start=start_datetime, end=end_datetime, freq=f"{time_resolution_minutes}min"
    )

    time_utc = time_utc.values.astype("datetime64[s]")
    time_utc = np.tile(time_utc, (batch_size, 1))
    return time_utc



"""Data pipeline for loading GSP national"""
import logging
from datetime import timedelta
from pathlib import Path
from typing import Union

import xarray as xr
from torch.utils.data.datapipes.datapipe import IterDataPipe

import ocf_datapipes  # noqa
from ocf_datapipes.batch import MergeNumpyModalities, MergeNWPNumpyModalities
from ocf_datapipes.config.load import load_yaml_configuration
from ocf_datapipes.config.model import Configuration
from ocf_datapipes.load import OpenGSPNational, OpenNWP
from ocf_datapipes.training.common import normalize_gsp
from ocf_datapipes.utils.consts import NWP_MEANS, NWP_STDS

logger = logging.getLogger(__name__)
xr.set_options(keep_attrs=True)

# should scale with batch_size #TODO
BUFFER_SIZE = 100


def gsp_national_datapipe(configuration_filename: Union[Path, str]) -> IterDataPipe:
    """
    Make GSP national data pipe

    Currently only has GSP and NWP's in them

    Args:
        configuration_filename: the configruation filename for the pipe

    Returns: datapipe
    """

    # Load configuration
    configuration: Configuration = load_yaml_configuration(configuration_filename)

    # Load and normalize GSP national data
    gsp_datapipe = OpenGSPNational(
        gsp_pv_power_zarr_path=configuration.input_data.gsp.gsp_zarr_path
    ).normalize(normalize_fn=normalize_gsp)

    # Load and normalize NWP data - There may be multiple NWP sources
    nwp_datapipe_dict = {}
    for nwp_source, nwp_conf in configuration.input_data.nwp.items():
        nwp_datapipe_dict[nwp_source] = OpenNWP(
            nwp_conf.nwp_zarr_path, provider=nwp_conf.nwp_provider
        ).normalize(mean=NWP_MEANS[nwp_conf.nwp_provider], std=NWP_STDS[nwp_conf.nwp_provider])

    # Add t0 idx to GSP
    gsp_datapipe = gsp_datapipe.add_t0_idx_and_sample_period_duration(
        sample_period_duration=timedelta(minutes=30),
        history_duration=timedelta(minutes=configuration.input_data.gsp.history_minutes),
    )

    # Add t0 idx to NWP and fork for time periods overlap
    nwp_time_periods_datapipes = {}
    for nwp_source, nwp_conf in configuration.input_data.nwp.items():
        nwp_datapipe_dict[nwp_source], nwp_time_periods_datapipes[nwp_source] = (
            nwp_datapipe_dict[nwp_source]
            .add_t0_idx_and_sample_period_duration(
                sample_period_duration=timedelta(hours=3),
                history_duration=timedelta(minutes=nwp_conf.history_minutes),
            )
            .fork(2)
        )

    # Fork the GSP data for different uses
    gsp_datapipe, gsp_time_periods_datapipe, gsp_valid_times_datapipe = gsp_datapipe.fork(3)

    # Get contiguous time periods for each data source
    gsp_time_periods_datapipe = gsp_time_periods_datapipe.find_contiguous_t0_time_periods(
        sample_period_duration=timedelta(minutes=30),
        history_duration=timedelta(minutes=configuration.input_data.gsp.history_minutes),
        forecast_duration=timedelta(minutes=configuration.input_data.gsp.forecast_minutes),
    )

    for nwp_source, nwp_conf in configuration.input_data.nwp.items():
        nwp_time_periods_datapipes[nwp_source] = nwp_time_periods_datapipes[
            nwp_source
        ].find_contiguous_t0_time_periods(
            sample_period_duration=timedelta(hours=3),
            history_duration=timedelta(minutes=nwp_conf.history_minutes),
            forecast_duration=timedelta(minutes=nwp_conf.forecast_minutes),
            time_dim="init_time_utc",
        )

    # Find joint overlapping time periods
    overlapping_datapipe = gsp_time_periods_datapipe.filter_to_overlapping_time_periods(
        secondary_datapipes=[*nwp_time_periods_datapipes.values()],
    )

    # Filter GSP to times valid for all data sources
    valid_periods_datapipe = gsp_valid_times_datapipe.filter_time_periods(
        time_periods=overlapping_datapipe
    )

    # Select t0 times
    gsp_t0_datapipe, nwp_t0_datapipe = valid_periods_datapipe.pick_t0_times().fork(2)

    # Take GSP time slices and convert to NumpyBatch
    gsp_numpy_datapipe = gsp_datapipe.select_time_slice(
        t0_datapipe=gsp_t0_datapipe,
        history_duration=timedelta(minutes=configuration.input_data.gsp.history_minutes),
        forecast_duration=timedelta(minutes=configuration.input_data.gsp.forecast_minutes),
        sample_period_duration=timedelta(minutes=30),
    ).convert_gsp_to_numpy_batch()

    # Take NWP time slices and convert to NumpyBatch
    nwp_numpy_modalities = dict()

    for nwp_source, nwp_conf in configuration.input_data.nwp.items():
        nwp_numpy_modalities[nwp_source] = (
            nwp_datapipe_dict[nwp_source]
            .select_time_slice_nwp(
                t0_datapipe=nwp_t0_datapipe,
                sample_period_duration=timedelta(hours=3),
                history_duration=timedelta(minutes=nwp_conf.history_minutes),
                forecast_duration=timedelta(minutes=nwp_conf.forecast_minutes),
            )
            .convert_nwp_to_numpy_batch()
        )

    # Combine the NWPs into NumpyBatch
    nwps_numpy_datapipe = MergeNWPNumpyModalities(nwp_numpy_modalities)

    # Join data sources together
    combined_datapipe = MergeNumpyModalities([gsp_numpy_datapipe, nwps_numpy_datapipe])

    # Now batch the data with batch size 4
    combined_datapipe = combined_datapipe.batch(4).merge_numpy_batch()

    return combined_datapipe



"""Compute the rolling mean of PV Power data"""
from typing import Optional, Union

import pandas as pd
import xarray as xr
from torch.utils.data import IterDataPipe, functional_datapipe


@functional_datapipe("pv_power_rolling_window")
class PVPowerRollingWindowIterDataPipe(IterDataPipe):
    """Compute rolling mean of PV power."""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        window: Union[int, pd.tseries.offsets.DateOffset, pd.core.indexers.objects.BaseIndexer] = 3,
        min_periods: Optional[int] = 2,
        center: bool = True,
        win_type: Optional[str] = None,
        expect_dataset: bool = True,
    ):
        """
        Compute the rolling mean of PV power data

        Args:
            source_datapipe: Datapipe emitting PV Xarray object

            window: Size of the moving window.
            If an integer, the fixed number of observations used for each window.

            If an offset, the time period of each window. Each window will be a variable sized
            based on the observations included in the time-period. This is only valid for
            datetimelike indexes. To learn more about the offsets & frequency strings, please see:
            https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases

            If a BaseIndexer subclass, the window boundaries based on the defined
            `get_window_bounds` method. Additional rolling keyword arguments,
            namely `min_periods` and `center` will be passed to `get_window_bounds`.

            min_periods: Minimum number of observations in window required to have a value;
            otherwise, result is `np.nan`.

            To avoid NaNs at the start and end of the timeseries, this should be <= ceil(window/2).

            For a window that is specified by an offset, `min_periods` will default to 1.

            For a window that is specified by an integer, `min_periods` will default to the size of
            the window.

            center: If False, set the window labels as the right edge of the window index.
            If True, set the window labels as the center of the window index.

            win_type: Window type
            expect_dataset: Whether to expect a dataset or DataArray
        """
        self.source_datapipe = source_datapipe
        self.window = window
        self.min_periods = min_periods
        self.center = center
        self.win_type = win_type
        self.expect_dataset = expect_dataset

    def __iter__(self) -> Union[xr.DataArray, xr.Dataset]:
        """Compute rolling mean of PV power"""
        for xr_data in self.source_datapipe:
            if self.expect_dataset:
                data_to_resample = xr_data["power_w"]
            else:
                data_to_resample = xr_data

            resampled = data_to_resample.rolling(
                dim={"time_utc": self.window},
                min_periods=self.min_periods,
                center=self.center,
            ).mean()

            if self.expect_dataset:
                xr_data["power_w"] = resampled
                resampled = xr_data

            # Resampling removes the attributes, so put them back:
            for attr_name in ("t0_idx", "sample_period_duration"):
                resampled.attrs[attr_name] = xr_data.attrs[attr_name]

            yield resampled



"""NWP Loader"""
import logging
from pathlib import Path
from typing import Union

import xarray as xr
from ocf_blosc2 import Blosc2  # noqa: F401
from torch.utils.data import IterDataPipe, functional_datapipe

from ocf_datapipes.load.nwp.providers.ecmwf import open_ifs
from ocf_datapipes.load.nwp.providers.gfs import open_gfs
from ocf_datapipes.load.nwp.providers.icon import open_icon_eu, open_icon_global
from ocf_datapipes.load.nwp.providers.ukv import open_ukv

logger = logging.getLogger(__name__)


@functional_datapipe("open_nwp")
class OpenNWPIterDataPipe(IterDataPipe):
    """Opens NWP Zarr and yields it"""

    def __init__(
        self,
        zarr_path: Union[Path, str, list[Path], list[str]],
        provider: str = "ukv",
    ):
        """
        Opens NWP Zarr and yields it

        Args:
            zarr_path: Path to the Zarr file
            provider: NWP provider
        """
        self.zarr_path = zarr_path
        logger.info(f"Using {provider.lower()}")
        if provider.lower() == "ukv":
            self.open_nwp = open_ukv
        elif provider.lower() == "icon-eu":
            self.open_nwp = open_icon_eu
        elif provider.lower() == "icon-global":
            self.open_nwp = open_icon_global
        elif provider.lower() == "ecmwf":
            self.open_nwp = open_ifs
        elif provider.lower() == "gfs":
            self.open_nwp = open_gfs
        else:
            raise ValueError(f"Unknown provider: {provider}")

    def __iter__(self) -> Union[xr.DataArray, xr.Dataset]:
        """Opens the NWP data"""
        logger.debug("Opening NWP data: %s", self.zarr_path)
        nwp = self.open_nwp(self.zarr_path)
        while True:
            yield nwp



"""Data pipeline for loading GSP, PV, Satellite and NWP"""
import logging
from datetime import timedelta
from pathlib import Path
from typing import Union

import xarray as xr
from torch.utils.data.datapipes.datapipe import IterDataPipe

import ocf_datapipes  # noqa
from ocf_datapipes.batch import MergeNumpyModalities, MergeNWPNumpyModalities
from ocf_datapipes.config.load import load_yaml_configuration
from ocf_datapipes.config.model import Configuration
from ocf_datapipes.load import OpenGSP, OpenNWP, OpenPVFromNetCDF, OpenSatellite
from ocf_datapipes.training.common import normalize_gsp, normalize_pv
from ocf_datapipes.utils.consts import NWP_MEANS, NWP_STDS, RSS_MEAN, RSS_STD

logger = logging.getLogger(__name__)
xr.set_options(keep_attrs=True)

# should scale with batch_size #TODO
BUFFER_SIZE = 100


def gsp_pv_nwp_satellite_data_pipeline(configuration: Union[Path, str]) -> IterDataPipe:
    """
    Make data pipe with GSP, PV, NWP and Satellite

    The locations are chosen by sampling from GSPs

    Args:
        configuration: the configuration filename for the pipe, can also be the actual configuration

    Returns: datapipe
    """

    # ----- LOAD AND NORMALIZE DATA

    # Load configuration
    configuration: Configuration = load_yaml_configuration(configuration)

    # Load and normalize GSP national data
    gsp_datapipe = OpenGSP(
        gsp_pv_power_zarr_path=configuration.input_data.gsp.gsp_zarr_path
    ).normalize(normalize_fn=normalize_gsp)

    # Load and normalize PV data, and fill night-time NaNs
    pv_datapipe = (
        OpenPVFromNetCDF(pv=configuration.input_data.pv)
        .pv_fill_night_nans()
        .normalize(normalize_fn=normalize_pv)
    )

    # Load and noralize satellite data
    satellite_datapipe = OpenSatellite(
        zarr_path=configuration.input_data.satellite.satellite_zarr_path
    ).normalize(mean=RSS_MEAN, std=RSS_STD)

    # Load and normalize NWP data - There may be multiple NWP sources
    nwp_datapipe_dict = {}
    for nwp_source, nwp_conf in configuration.input_data.nwp.items():
        nwp_datapipe_dict[nwp_source] = OpenNWP(
            nwp_conf.nwp_zarr_path, provider=nwp_conf.nwp_provider
        ).normalize(mean=NWP_MEANS[nwp_conf.nwp_provider], std=NWP_STDS[nwp_conf.nwp_provider])

    # ----- ADD t0 IDX

    # Add t0 idx to GSP
    gsp_datapipe = gsp_datapipe.add_t0_idx_and_sample_period_duration(
        sample_period_duration=timedelta(
            minutes=configuration.input_data.gsp.time_resolution_minutes
        ),
        history_duration=timedelta(minutes=configuration.input_data.gsp.history_minutes),
    )

    # Add t0 idx to PV
    pv_datapipe = pv_datapipe.add_t0_idx_and_sample_period_duration(
        sample_period_duration=timedelta(
            minutes=configuration.input_data.pv.time_resolution_minutes
        ),
        history_duration=timedelta(minutes=configuration.input_data.pv.history_minutes),
    )

    # Add t0 idx to satellite
    satellite_datapipe = satellite_datapipe.add_t0_idx_and_sample_period_duration(
        sample_period_duration=timedelta(
            minutes=configuration.input_data.satellite.time_resolution_minutes
        ),
        history_duration=timedelta(minutes=configuration.input_data.satellite.history_minutes),
    )

    # Add t0 idx to NWP and fork for time periods overlap
    nwp_time_periods_datapipes = {}
    for nwp_source in nwp_datapipe_dict.keys():
        nwp_datapipe_dict[nwp_source], nwp_time_periods_datapipes[nwp_source] = (
            nwp_datapipe_dict[nwp_source]
            .add_t0_idx_and_sample_period_duration(
                sample_period_duration=timedelta(hours=3),
                history_duration=timedelta(
                    minutes=configuration.input_data.nwp[nwp_source].history_minutes
                ),
            )
            .fork(2)
        )

    # ----- CORE DATASET FORK

    # Fork the GSP data for different uses
    (
        gsp_datapipe,
        gsp_time_periods_datapipe,
        gsp_valid_times_datapipe,
        gsp_location_datapipe,
    ) = gsp_datapipe.fork(4)

    # ----- LOAD AND NORMALIZE DATA

    # Pick locations
    location_datapipe = gsp_location_datapipe.pick_locations()

    # Take PV space slice
    loc_dp, location_datapipe = location_datapipe.fork(2, buffer_size=BUFFER_SIZE)

    pv_datapipe = pv_datapipe.select_spatial_slice_meters(
        location_datapipe=loc_dp,
        roi_height_meters=configuration.input_data.pv.pv_image_size_meters_height,
        roi_width_meters=configuration.input_data.pv.pv_image_size_meters_width,
        dim_name="pv_system_id",
    )

    # Take satellite space slice
    loc_dp, location_datapipe = location_datapipe.fork(2, buffer_size=BUFFER_SIZE)

    satellite_datapipe = satellite_datapipe.select_spatial_slice_pixels(
        location_datapipe=loc_dp,
        roi_height_pixels=configuration.input_data.satellite.satellite_image_size_pixels_height,
        roi_width_pixels=configuration.input_data.satellite.satellite_image_size_pixels_width,
    )

    # Take NWP space slice
    nwp_numpy_modalities = dict()

    for nwp_source, nwp_conf in configuration.input_data.nwp.items():
        loc_dp, location_datapipe = location_datapipe.fork(2, buffer_size=BUFFER_SIZE)

        nwp_datapipe_dict[nwp_source] = nwp_datapipe_dict[nwp_source].select_spatial_slice_pixels(
            loc_dp,
            roi_height_pixels=nwp_conf.nwp_image_size_pixels_height,
            roi_width_pixels=nwp_conf.nwp_image_size_pixels_width,
        )

    # Take GSP space slice
    gsp_datapipe = gsp_datapipe.select_spatial_slice_meters(
        location_datapipe=location_datapipe,
        roi_height_meters=configuration.input_data.gsp.gsp_image_size_pixels_height,
        roi_width_meters=configuration.input_data.gsp.gsp_image_size_pixels_width,
        dim_name="gsp_id",
    )

    # ----- SELECT TIME SLICES

    # GSP get contiguous time periods
    gsp_time_periods_datapipe = gsp_time_periods_datapipe.find_contiguous_t0_time_periods(
        sample_period_duration=timedelta(minutes=30),
        history_duration=timedelta(minutes=configuration.input_data.gsp.history_minutes),
        forecast_duration=timedelta(minutes=configuration.input_data.gsp.forecast_minutes),
    )

    # PV get contiguous time periods
    pv_datapipe, pv_time_periods_datapipe = pv_datapipe.fork(2)

    pv_time_periods_datapipe = pv_time_periods_datapipe.find_contiguous_t0_time_periods(
        sample_period_duration=timedelta(minutes=5),
        history_duration=timedelta(minutes=configuration.input_data.pv.history_minutes),
        forecast_duration=timedelta(minutes=configuration.input_data.pv.forecast_minutes),
    )

    # Satellite get contiguous time periods
    satellite_datapipe, satellite_time_periods_datapipe = satellite_datapipe.fork(2)

    satellite_time_periods_datapipe = (
        satellite_time_periods_datapipe.find_contiguous_t0_time_periods(
            sample_period_duration=timedelta(
                minutes=configuration.input_data.satellite.time_resolution_minutes
            ),
            history_duration=timedelta(minutes=configuration.input_data.satellite.history_minutes),
            forecast_duration=timedelta(
                minutes=configuration.input_data.satellite.forecast_minutes
            ),
        )
    )

    # NWP get contiguous time periods
    for nwp_source, nwp_conf in configuration.input_data.nwp.items():
        nwp_time_periods_datapipes[nwp_source] = nwp_time_periods_datapipes[
            nwp_source
        ].find_contiguous_t0_time_periods(
            sample_period_duration=timedelta(hours=3),
            history_duration=timedelta(minutes=nwp_conf.history_minutes),
            forecast_duration=timedelta(minutes=nwp_conf.forecast_minutes),
            time_dim="init_time_utc",
        )

    # Find joint overlapping time periods
    overlapping_datapipe = gsp_time_periods_datapipe.filter_to_overlapping_time_periods(
        secondary_datapipes=[
            pv_time_periods_datapipe,
            satellite_time_periods_datapipe,
            *nwp_time_periods_datapipes.values(),
        ],
    )

    # Filter to times valid for all data sources
    valid_periods_datapipe = gsp_valid_times_datapipe.filter_time_periods(
        time_periods=overlapping_datapipe
    )

    # Select t0 times
    t0_datapipe = valid_periods_datapipe.pick_t0_times()

    # Take NWP time slices and convert to NumpyBatch
    nwp_numpy_modalities = dict()

    for nwp_source, nwp_conf in configuration.input_data.nwp.items():
        nwp_t0_datapipe, t0_datapipe = t0_datapipe.fork(2)

        nwp_numpy_modalities[nwp_source] = (
            nwp_datapipe_dict[nwp_source]
            .select_time_slice_nwp(
                t0_datapipe=nwp_t0_datapipe,
                sample_period_duration=timedelta(hours=3),
                history_duration=timedelta(minutes=nwp_conf.history_minutes),
                forecast_duration=timedelta(minutes=nwp_conf.forecast_minutes),
            )
            .convert_nwp_to_numpy_batch()
        )

    # Combine the NWPs into NumpyBatch
    nwps_numpy_datapipe = MergeNWPNumpyModalities(nwp_numpy_modalities)

    # Take PV time slices and convert to NumpyBatch
    pv_t0_datapipe, t0_datapipe = t0_datapipe.fork(2)

    pv_numpy_datapipe = pv_datapipe.select_time_slice(
        t0_datapipe=pv_t0_datapipe,
        sample_period_duration=timedelta(minutes=5),
        history_duration=timedelta(minutes=configuration.input_data.pv.history_minutes),
        forecast_duration=timedelta(minutes=configuration.input_data.pv.forecast_minutes),
    ).convert_pv_to_numpy_batch()

    # Take satellite time slices and convert to NumpyBatch
    satellite_t0_datapipe, t0_datapipe = t0_datapipe.fork(2)

    satellite_numpy_datapipe = satellite_datapipe.select_time_slice(
        t0_datapipe=satellite_t0_datapipe,
        sample_period_duration=timedelta(
            minutes=configuration.input_data.satellite.time_resolution_minutes
        ),
        history_duration=timedelta(minutes=configuration.input_data.satellite.history_minutes),
        forecast_duration=timedelta(minutes=configuration.input_data.satellite.forecast_minutes),
    ).convert_satellite_to_numpy_batch()

    # Take GSP time slices and convert to NumpyBatch
    gsp_numpy_datapipe = gsp_datapipe.select_time_slice(
        t0_datapipe=t0_datapipe,
        history_duration=timedelta(minutes=configuration.input_data.gsp.history_minutes),
        forecast_duration=timedelta(minutes=configuration.input_data.gsp.forecast_minutes),
        sample_period_duration=timedelta(minutes=30),
    ).convert_gsp_to_numpy_batch()

    # Join datapipes together
    combined_datapipe = MergeNumpyModalities(
        [gsp_numpy_datapipe, pv_numpy_datapipe, satellite_numpy_datapipe, nwps_numpy_datapipe]
    )

    # Now batch the data with batch size 4
    combined_datapipe = combined_datapipe.batch(4).merge_numpy_batch()

    return combined_datapipe



"""Datapipe to check and optionally fill NaNs in data"""
from typing import Union

import numpy as np
import xarray as xr
from torch.utils.data import IterDataPipe, functional_datapipe


@functional_datapipe("check_nans")
class CheckNaNsIterDataPipe(IterDataPipe):
    """Checks, and optionally fills, NaNs in Xarray Dataset"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        dataset_name: str = None,
        fill_nans: bool = False,
        fill_value: float = 0.0,
    ):
        """
        Checks and optionally fills NaNs in the data

        Args:
            source_datapipe: Datapipe emitting Xarray Datasets
            dataset_name: Optional name for dataset to check, if None, checks whole dataset
            fill_nans: Whether to fill NaNs with 0 or not
            fill_value: Value to fill NaNs with
        """
        self.source_datapipe = source_datapipe
        self.dataset_name = dataset_name
        self.fill_nans = fill_nans
        self.source_datapipe_name = source_datapipe.__repr__()
        self.fill_value = fill_value

    def __iter__(self) -> Union[xr.DataArray, xr.Dataset]:
        """
        Checks for NaNs in data

        Returns:
            Original Xarray Dataset, after validation, and potentially with NaNs filled
        """
        for xr_data in self.source_datapipe:
            if self.fill_nans:
                if self.dataset_name is None:
                    xr_data = self.fill_nan(data=xr_data, fill_value=self.fill_value)
                else:
                    xr_data[self.dataset_name] = self.fill_nan(
                        data=xr_data[self.dataset_name],
                        fill_value=self.fill_value,
                    )
            self.check_nan_and_inf(
                data=xr_data if self.dataset_name is None else xr_data[self.dataset_name],
            )
            yield xr_data

    def check_nan_and_inf(self, data: xr.Dataset) -> None:
        """Check that all values are non NaNs and not infinite"""

        if np.isnan(data).any():
            if self.dataset_name is None:
                message = f"Some data values are NaNs in datapipe {self.datapipe_name}. "
            else:
                message = (
                    f"Some data values are NaNs in datapipe {self.datapipe_name},"
                    f" dataset {self.dataset_name}. "
                )

            # find out which example has nans in it
            for i in range(data.shape[0]):
                if np.isnan(data[i]).any():
                    message += f" Nans in example {i}."
            raise Warning(message)

        if np.isinf(data).any():
            message = f"Some data values are Infinite in datapipe {self.datapipe_name}."
            raise Warning(message)

    def fill_nan(self, data: xr.Dataset, fill_value: float = 0.0) -> xr.Dataset:
        """Check that all values are non NaNs and not infinite"""

        if np.isnan(data).any():
            data = data.fillna(fill_value)

        return data



import logging
from timeit import timeit

import numpy as np
import pandas as pd
import xarray as xr

from ocf_datapipes.transform.xarray import AssignDayNightStatus

logger = logging.getLogger(__name__)


def test_assign_status_night(passiv_datapipe):
    night_status = AssignDayNightStatus(passiv_datapipe)
    data = next(iter(night_status))
    data_coords = data.coords["status_daynight"].values
    # For the fourth month in the passiv_datapipe, according to
    # the uk daynight dictionary set in the assign_daynight_status.py
    # total night status count of 5 minute timseries data is 121
    assert np.count_nonzero(data_coords == "night") == 121.0
    assert "day" and "night" in data.coords["status_daynight"].values


def test_time(passiv_datapipe):
    # Create the instance of the AssignDayNightStatusIterDataPipe class
    datapipe = AssignDayNightStatus(passiv_datapipe)

    # Using timeit to measure the execution time of the __iter__ method
    # number of simulations takes place are 365 (meaning for a year),
    # number of simulations takes place are 365 (meaning for a year),
    # as the datapipe considers one day worth of data
    execution_time = timeit(lambda: next(iter((datapipe))), number=365)

    logger.info(f"Execution time to test for 365 times:\n{execution_time:.4f} seconds")
    print(f"\nExecution time to test for 365 times: {execution_time:.4f} seconds")


def test_with_constructed_array():
    time = pd.date_range("2022-01-01 17:00", "2022-01-01 23:55", freq="5min")
    pv_system_id = [1, 2, 3]
    ALL_COORDS = {"time_utc": time, "pv_system_id": pv_system_id}

    data = np.zeros((len(time), len(pv_system_id)))
    data[:, 2] = 1.0

    data_array = xr.DataArray(
        data,
        coords=ALL_COORDS,
    )

    night_status = AssignDayNightStatus([data_array])
    data = next(iter(night_status))
    data_coords = data.coords["status_daynight"].values

    # As the time range only includes night timestamps
    # status "day" is not assigned to any of the timestamps
    assert "day" not in data_coords



"""Configuration Loader"""
import logging

from torch.utils.data.datapipes._decorator import functional_datapipe
from torch.utils.data.datapipes.datapipe import IterDataPipe

from ocf_datapipes.config.load import load_yaml_configuration

logger = logging.getLogger(__name__)


@functional_datapipe("open_config")
class OpenConfigurationIterDataPipe(IterDataPipe):
    """Open and return the configuration data"""

    def __init__(self, configuration_filename: str):
        """
        Open and return config data

        Args:
            configuration_filename: Filename to open
        """
        self.configuration_filename = configuration_filename

    def __iter__(self):
        """Open and return configuration file"""
        logger.debug(f"Going to open {self.configuration_filename}")

        configuration = load_yaml_configuration(self.configuration_filename)

        while True:
            yield configuration



""" Make fake Satellite data """
from datetime import datetime

import numpy as np

from ocf_datapipes.batch import BatchKey
from ocf_datapipes.batch.fake.utils import get_n_time_steps_from_config, make_time_utc
from ocf_datapipes.config.model import Configuration


def make_fake_satellite_data(
    configuration: Configuration,
    t0_datetime_utc: datetime,
    is_hrv: bool = False,
    batch_size: int = 8,
) -> dict:
    """
    Make Fake Satellite data ready for ML model. This makes data across all different data inputs

    Args:
        configuration: configuration object
        t0_datetime_utc: one datetime for when t0 is
        is_hrv: option if its hrv or not
        batch_size: Integer batch size to create

    Returns: dictionary of satellite items
    """

    if is_hrv:
        prefix = "hrv"
    else:
        prefix = ""

    variable = prefix + "satellite"

    satellite_config = getattr(configuration.input_data, variable)

    if satellite_config is None:
        return {}

    n_channels = len(getattr(satellite_config, f"{variable}_channels"))
    height = getattr(satellite_config, f"{variable}_image_size_pixels_height")
    width = getattr(satellite_config, f"{variable}_image_size_pixels_width")
    n_fourier_features = 8

    # make time matrix
    time_utc = make_time_utc(
        batch_size=batch_size,
        history_minutes=satellite_config.history_minutes,
        forecast_minutes=satellite_config.forecast_minutes,
        t0_datetime_utc=t0_datetime_utc,
        time_resolution_minutes=satellite_config.time_resolution_minutes,
    )
    n_times = time_utc.shape[1]

    batch = {}
    # shape: (batch_size, time, channels, y, x)
    #
    # Or, if the imagery has been patched,
    # shape: (batch_size, time, channels, y, x, n_pixels_per_patch) where n_pixels_per_patch
    # is the *total* number of pixels, # TODO
    batch[getattr(BatchKey, f"{variable}_actual")] = np.random.random(
        (batch_size, n_times, n_channels, height, width)
    )
    batch[getattr(BatchKey, f"{variable}_predicted")] = np.random.random(
        (batch_size, n_times, n_channels, height, width)
    )

    batch[getattr(BatchKey, f"{variable}_t0_idx")] = get_n_time_steps_from_config(
        satellite_config, include_forecast=False
    )

    # HRV satellite coordinates:
    batch[getattr(BatchKey, f"{variable}_y_osgb")] = np.random.randint(0, 100, (batch_size, height))
    batch[getattr(BatchKey, f"{variable}_x_osgb")] = np.random.randint(0, 100, (batch_size, width))
    batch[getattr(BatchKey, f"{variable}_y_geostationary")] = np.random.randint(
        0, 100, (batch_size, height)
    )
    batch[getattr(BatchKey, f"{variable}_x_geostationary")] = np.random.randint(
        0, 100, (batch_size, width)
    )

    batch[
        getattr(BatchKey, f"{variable}_time_utc")
    ] = time_utc  # Seconds since UNIX epoch (1970-01-01).
    # Added by np_batch_processor.Topography:
    batch[getattr(BatchKey, f"{variable}_surface_height")] = np.random.randint(
        0, 100, (batch_size, height, width)
    )

    # HRV satellite Fourier coordinates:
    # Spatial coordinates. Shape: (batch_size, y, x, n_fourier_features_per_dim)
    batch[getattr(BatchKey, f"{variable}_y_osgb_fourier")] = np.random.random(
        (batch_size, height, n_fourier_features)
    )  # TODO check its just 3dims
    batch[getattr(BatchKey, f"{variable}_x_osgb_fourier")] = np.random.random(
        (batch_size, width, n_fourier_features)
    )

    #: Time shape: (batch_size, n_timesteps, n_fourier_features_per_dim)
    batch[getattr(BatchKey, f"{variable}_time_utc_fourier")] = np.random.random(
        (batch_size, n_times, n_fourier_features)
    )

    return batch



import numpy as np

from ocf_datapipes.transform.xarray import Normalize
from ocf_datapipes.utils.consts import NWP_MEANS, NWP_STDS, RSS_MEAN, RSS_STD
import pytest


def test_normalize_sat(sat_datapipe):
    normed_sat_datapipe = Normalize(
        sat_datapipe,
        # HRV dosn't have channel dimension
        mean=RSS_MEAN.sel(channel="HRV"),
        std=RSS_STD.sel(channel="HRV"),
    )
    data = next(iter(normed_sat_datapipe))
    assert np.all(data <= 10.0)
    assert np.all(data >= -10.0)


def test_normalize_nwp(nwp_datapipe):
    normed_nwp_datapipe = Normalize(nwp_datapipe, mean=NWP_MEANS["ukv"], std=NWP_STDS["ukv"])
    data = next(iter(normed_nwp_datapipe))
    assert np.all(data.values <= 10.0)
    assert np.all(data.values >= -10.0)


def test_normalize_topo(topo_datapipe):
    normed_topo_datapipe = topo_datapipe.reproject_topography().normalize(
        calculate_mean_std_from_example=True
    )
    data = next(iter(normed_topo_datapipe))
    assert data.mean().compute() == pytest.approx(0, abs=0.001)
    assert data.std().compute() == pytest.approx(1, abs=0.001)


def test_normalize_gsp(gsp_datapipe):
    normed_gsp_datapipe = gsp_datapipe.normalize(
        normalize_fn=lambda x: x / x.effective_capacity_mwp
    )
    data = next(iter(normed_gsp_datapipe))
    assert np.min(data) >= 0.0
    assert np.max(data) <= 1.0


def test_normalize_passiv(passiv_datapipe):
    normed_passiv_datapipe = passiv_datapipe.normalize(
        normalize_fn=lambda x: x / x.observed_capacity_wp
    )
    data = next(iter(normed_passiv_datapipe))
    assert np.min(data) >= 0.0
    assert np.max(data) <= 1.0


def test_normalize_pvoutput(pvoutput_datapipe):
    normed_pvoutput_datapipe = pvoutput_datapipe.normalize(
        normalize_fn=lambda x: x / x.observed_capacity_wp
    )
    data = next(iter(normed_pvoutput_datapipe))
    assert np.min(data) >= 0.0
    assert np.max(data) <= 1.0
